/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gmime/gmime.h>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/log.h>
#include <pan/general/i18n.h>
#include <pan/usenet-utils/gnksa.h>
#include <pan/usenet-utils/mime-utils.h>
#include <pan/usenet-utils/utf8-utils.h>
#include <pan/usenet-utils/url-find.h>
#include <pan/icons/pan-pixbufs.h>
#include "body-pane.h"
#include "pad.h"
#include "xface.h"
#include "url.h"

#define FIRST_PICTURE "first-picture"

using namespace pan;

/***
****  URL
***/
namespace
{
  char* get_url_from_location (GtkWidget * w, int x, int y)
  {
    char * retval (0);
    static GtkTextTag * url_tag = NULL;

    // get the buffer
    g_return_val_if_fail (GTK_IS_TEXT_VIEW(w), FALSE);
    GtkTextBuffer * text_buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW(w));

    // translate coordinates
    int old_x (x);
    int old_y (y);
    gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW(w), GTK_TEXT_WINDOW_TEXT, old_x, old_y, &x, &y);

    // did the user click on a url?
    GtkTextIter pos;
    gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW(w), &pos, x, y);
    if (!url_tag) {
      GtkTextTagTable * tag_table = gtk_text_buffer_get_tag_table (text_buffer);
      url_tag = gtk_text_tag_table_lookup (tag_table, "url");
      if (!url_tag)
        return 0;
    }

    const bool clicked_on_url = gtk_text_iter_has_tag (&pos, url_tag);
    if (clicked_on_url) { // get the url
      GtkTextIter begin=pos, end=pos;
      gtk_text_iter_backward_to_tag_toggle (&begin, NULL);
      gtk_text_iter_forward_to_tag_toggle (&end, NULL);
      retval = gtk_text_iter_get_text (&begin, &end);
    }

    return retval;
  }

  gboolean text_button_pressed (GtkWidget * w, GdkEventButton * event, gpointer unused)
  {
    g_return_val_if_fail (GTK_IS_TEXT_VIEW(w), FALSE);

    if (event->button==1 || event->button==2) {
      char * url = get_url_from_location (w, (int)event->x, (int)event->y);
      if (url) {
        /* this is kind of a crude way of making sure that double-click
         * doesn't open two or three browser windows. */
        static time_t last_url_time = 0;
        const time_t this_url_time = time (NULL);
        if (this_url_time != last_url_time) {
          last_url_time = this_url_time;
          pan_url_show (url);
          g_free (url);
        }
      }
    }
    return false;
  }

  gboolean motion_notify_event (GtkWidget * w, GdkEventMotion * event, gpointer user_data)
  {
    static GdkCursor * cursor_current (0);
    static GdkCursor * cursor_ibeam (0);
    static GdkCursor * cursor_href (0);

    if (event->window != NULL)
    {
      // initialize static variables
      if (!cursor_ibeam)
        cursor_ibeam = gdk_cursor_new (GDK_XTERM);
      if (!cursor_href)
        cursor_href = gdk_cursor_new (GDK_HAND2);

      // pump out x, y, and state
      int x, y;
      GdkModifierType state;
      if (event->is_hint)
        gdk_window_get_pointer (event->window, &x, &y, &state);
      else {
        x = (int)event->x;
        y = (int)event->y;
        state = (GdkModifierType)event->state;
      }

      // decide what cursor we should be using
      GdkCursor * cursor_new (cursor_ibeam);
      char * url = get_url_from_location (w, (int)event->x, (int)event->y);
      if (url) {
        cursor_new = cursor_href;
        g_free (url);
      }

      // change the cursor if needed
      if (cursor_new != cursor_current)
        gdk_window_set_cursor (event->window, cursor_current=cursor_new);
    }

    return false;
  }
}

/***
****  INIT
***/
namespace
{
  GtkTextTag* get_tag (GtkTextTagTable * table, const char * key)
  {
    g_assert (table);
    g_assert (key && *key);

    GtkTextTag * tag = gtk_text_tag_table_lookup (table, key);
    if (!tag) {
      tag = gtk_text_tag_new (key);
      gtk_text_tag_table_add (table, tag);
      g_object_unref (tag); // table refs it
    }
    return tag;
  }

  void
  set_text_buffer_tags (GtkTextBuffer * buffer, const Prefs& prefs)
  {
    GtkTextTagTable * table = gtk_text_buffer_get_tag_table (buffer);

    g_object_set (get_tag(table,"bold"),
                  "weight", PANGO_WEIGHT_BOLD,
                  NULL);
    
    g_object_set (get_tag (table, "italic"),
                  "style", PANGO_STYLE_ITALIC,
                  NULL);

    g_object_set (get_tag (table, "underline"),
                  "underline", PANGO_UNDERLINE_SINGLE,
                  NULL);

    g_object_set (get_tag (table, "url"),
                  "underline", PANGO_UNDERLINE_SINGLE,
                  "foreground", prefs.get_color_str ("body-pane-color-url", "#0000ff").c_str(),
                  NULL);

    get_tag (table, "quote_0");

    g_object_set (get_tag (table, "quote_1"),
                  "foreground", prefs.get_color_str ("body-pane-color-quote-1", "#550077").c_str(),
                  NULL);

    g_object_set (get_tag (table, "quote_2"),
                  "foreground", prefs.get_color_str ("body-pane-color-quote-2", "#006600").c_str(),
                  NULL);

    g_object_set (get_tag (table, "quote_3"),
                  "foreground", prefs.get_color_str ("body-pane-color-quote-3", "#bb5500").c_str(),
                  NULL);

    g_object_set (get_tag (table, "signature"),
                  "foreground", prefs.get_color_str ("body-pane-color-signature", "#00AAAA").c_str(),
                  NULL);
  }
}

/***
****
***/
namespace
{
  // handle up, down, pgup, pgdown to scroll
  gboolean text_key_pressed (GtkWidget * w, GdkEventKey * event, gpointer scroll)
  {
    gboolean handled (false);

    g_return_val_if_fail (GTK_IS_TEXT_VIEW(w), FALSE);
    const bool up = event->keyval==GDK_Up || event->keyval==GDK_KP_Up;
    const bool down = event->keyval==GDK_Down || event->keyval==GDK_KP_Down;

    if (up || down)
    {
      handled = true;
      gtk_text_view_place_cursor_onscreen (GTK_TEXT_VIEW(w));
      GtkAdjustment * adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW(scroll));
      gdouble val = adj->value;
      if (up)
        val -= adj->step_increment;
      else
        val += adj->step_increment;
      val = MAX(val, adj->lower);
      val = MIN(val, adj->upper-adj->page_size);
      gtk_adjustment_set_value (adj, val);
    }

    return handled;
  }
}

/****
*****
*****   SETTING THE TEXT FROM AN ARTICLE
*****
****/
namespace
{
  bool text_get_show_all_headers () { return false; }

  /**
   * Returns the quote tag ("quote_0", "quote_1", etc.) appropriate for the line.
   * The right tag is calculated by adding up the number of quote characters
   * at the beginning of the line.
   *
   * @param utf8_line the line whose quote status we're checking
   * @param utf8_byte_len the byte length of utf8_line
   * @return a const string for the line's quote tag.  Never NULL.
   */
  const char * get_quote_tag (const TextMassager     * text_massager,
                              const char             * utf8_line,
                              int                      utf8_byte_len)
  {
    const char * str = utf8_line;
    const char * line_end = utf8_line + utf8_byte_len;
    const char * retval = "quote_0";

    if (0<utf8_byte_len && str && *str)
    {
      int depth = 0;

      /* walk past leading spaces */
      while (str!=line_end && g_unichar_isspace(g_utf8_get_char(str)))
        str = g_utf8_next_char (str);

      /* count the number of spaces or quote characters */
      for (;;) {
        if (str == line_end)
          break;
        else if (text_massager->is_quote_character (g_utf8_get_char (str))) 
          ++depth;
        else if (!g_unichar_isspace(g_utf8_get_char(str)))
          break;
        str = g_utf8_next_char (str);
      }

      if (!depth)
        retval = "quote_0";
      else switch (depth % 3) {
        case 1: retval = "quote_1"; break;
        case 2: retval = "quote_2"; break;
        case 0: retval = "quote_3"; break;
      }
    }

    return retval;
  }

  GdkPixbuf*
  get_emoticon (const std::string& key)
  {
    typedef std::map<std::string,GdkPixbuf*> pixbufs_t;
    static pixbufs_t pixbufs;
    static bool inited (false);
    if (!inited) {
      inited = true;
      pixbufs[":)"] = gdk_pixbuf_new_from_inline (-1, icon_mozilla_smile, false, 0);
      pixbufs[";)"] = gdk_pixbuf_new_from_inline (-1, icon_mozilla_wink, false, 0);
      pixbufs[":("] = gdk_pixbuf_new_from_inline (-1, icon_mozilla_frown, false, 0);
      pixbufs[":P"] = gdk_pixbuf_new_from_inline (-1, icon_mozilla_tongueout, false, 0);
      pixbufs[":O"] = gdk_pixbuf_new_from_inline (-1, icon_mozilla_surprised, false, 0);
    }
    pixbufs_t::const_iterator it (pixbufs.find (key));
    return it==pixbufs.end() ? 0 : it->second;
  }

  void
  set_section_tag (GtkTextBuffer * buffer,
                   GtkTextIter * start,
                   const StringView& body,
                   const StringView& area,
                   const char * tag)
  {
    // if no alnums, chances are area is a false positive
    int alnums (0);
    for (const char *pch(area.begin()), *end(area.end()); pch!=end; ++pch)
      if (::isalnum(*pch))
        ++alnums;
    if (!alnums)
      return;

    GtkTextIter mark_start = *start;
    gtk_text_iter_forward_chars (&mark_start, g_utf8_strlen(body.str, area.str-body.str));
    GtkTextIter mark_end = mark_start;
    gtk_text_iter_forward_chars (&mark_end, g_utf8_strlen(area.str,area.len));
    //gtk_text_buffer_remove_all_tags (buffer, &mark_start, &mark_end);
    gtk_text_buffer_apply_tag_by_name (buffer, tag, &mark_start, &mark_end);
  }

  void
  replace_emoticon_text_with_pixbuf (GtkTextBuffer      * buffer,
                                     GtkTextMark        * mark,
                                     std::string        & body,
                                     const std::string  & key)
  {
    GdkPixbuf * pixbuf = get_emoticon (key);
    g_assert (pixbuf != 0);
    std::string::size_type pos;
    while (((pos=body.find(key))) != std::string::npos)
    {
      GtkTextIter begin;
      gtk_text_buffer_get_iter_at_mark (buffer, &begin, mark);
      gtk_text_iter_forward_chars (&begin, g_utf8_strlen(&body[0],pos));
      GtkTextIter end = begin;
      gtk_text_iter_forward_chars (&end, key.size());
      gtk_text_buffer_delete (buffer, &begin, &end);
      body.erase (pos, key.size());
      gtk_text_buffer_insert_pixbuf (buffer, &end, pixbuf);
      body.insert (pos, 1, '?'); // make the body.size() match the textbuf's size
    }
  }

  /**
   * Appends the specified body into the text buffer.
   * This function takes care of muting quotes and marking
   * quoted and URL areas in the GtkTextBuffer.
   */
  void
  append_text_buffer_nolock (const TextMassager  * text_massager,
                             GtkTextBuffer       * buffer,
                             const char          * body_pch,
                             bool                  mute_quotes,
                             bool                  show_smilies,
                             bool                  do_markup)
  {
    g_return_if_fail (buffer!=0);
    g_return_if_fail (GTK_IS_TEXT_BUFFER(buffer));


    // mute the quoted text, if desired
    if (!body_pch) body_pch = "";
    std::string body (mute_quotes ? text_massager->mute_quotes (body_pch) : body_pch);
    //char * pan_find_signature_delimiter (body, &sig_point);

    // insert the text
    GtkTextIter end;
    gtk_text_buffer_get_end_iter (buffer, &end);
    GtkTextMark * mark = gtk_text_buffer_create_mark (buffer, "blah", &end, TRUE);
    gtk_text_buffer_insert (buffer, &end, body.c_str(), -1);
    GtkTextIter start;
    gtk_text_buffer_get_iter_at_mark (buffer, &start, mark);

    StringView v(body), line;
    GtkTextIter mark_start = start;

    // find where the signature begins...
    const char * sig_point (0);
    int offset (0);
    if (GNKSA::find_signature_delimiter (v, offset) != GNKSA::SIG_NONE)
      sig_point = v.str + offset;
   
    // colorize the quoted text 
    GtkTextIter mark_end;
    std::string last_quote_tag;
    bool is_sig (false);
    const char * last_quote_begin (v.str);
    while (v.pop_token (line, '\n'))
    {
      if (line.empty())
        continue;

      if (line.str == sig_point)
        is_sig = true;

      const std::string quote_tag = is_sig
        ?  "signature"
        : get_quote_tag (text_massager, line.str, line.len);

      // if we've changed tags, colorize the previous block
      if (!last_quote_tag.empty() && quote_tag!=last_quote_tag) {
        mark_end = mark_start;
        gtk_text_iter_forward_chars (&mark_end, g_utf8_strlen(last_quote_begin,line.str-1-last_quote_begin));
        gtk_text_buffer_apply_tag_by_name (buffer, last_quote_tag.c_str(), &mark_start, &mark_end);
        mark_start = mark_end;
        gtk_text_iter_forward_chars (&mark_start, 1);
        last_quote_begin = line.str;
      }

      last_quote_tag = quote_tag;
    }

    // apply the final tag, if there is one */
    if (!last_quote_tag.empty()) {
      gtk_text_buffer_get_end_iter (buffer, &mark_end);
      gtk_text_buffer_apply_tag_by_name (buffer, last_quote_tag.c_str(), &mark_start, &mark_end);
    }

    const StringView v_all (body);
    if (do_markup) {
      // find boldface, italic, and underline
      const char * mode_begin (0);
      const char * b (v_all.begin());
      const char * e (v_all.end());
      enum { NORMAL, UNDERLINE, BOLD, ITALIC };
      int mode = NORMAL;
      for (const char * pch(b); pch!=e; ++pch) {
        if (*pch=='_' && mode==UNDERLINE && (pch+1==e || ::isspace(pch[1]) || ::ispunct(pch[1])))
          set_section_tag (buffer, &start, v_all, StringView(mode_begin,pch+1-mode_begin), "underline");
        if (*pch=='/' && mode==ITALIC && (pch+1==e || ::isspace(pch[1]) || ::ispunct(pch[1])))
          set_section_tag (buffer, &start, v_all, StringView(mode_begin,pch+1-mode_begin), "italic");
        if (*pch=='*' && mode==BOLD && (pch+1==e || ::isspace(pch[1]) || ::ispunct(pch[1])))
          set_section_tag (buffer, &start, v_all, StringView(mode_begin,pch+1-mode_begin), "bold");
        else if (*pch=='_' && (pch==b ||isspace(pch[-1])))
          { mode = UNDERLINE; mode_begin=pch; }
        else if (*pch=='*' && (pch==b ||isspace(pch[-1])))
          { mode = BOLD; mode_begin=pch; }
        else if (*pch=='/' && (pch==b ||isspace(pch[-1])))
          { mode = ITALIC; mode_begin=pch; }
        else if (isspace(*pch))
          { mode = NORMAL; mode_begin=pch; }
      }
    }
    
    // colorize urls
    StringView area;
    StringView march (v_all);
    while ((url_find (march, area))) {
      set_section_tag (buffer, &start, v_all, area, "url");
      march = march.substr (area.str + area.len, 0);
    }

    // do this last, since it actually mutates the text instead of just marking it up
    if (show_smilies) {
      replace_emoticon_text_with_pixbuf (buffer, mark, body, ":)");
      replace_emoticon_text_with_pixbuf (buffer, mark, body, ";)");
      replace_emoticon_text_with_pixbuf (buffer, mark, body, ":O");
      replace_emoticon_text_with_pixbuf (buffer, mark, body, ":P");
      replace_emoticon_text_with_pixbuf (buffer, mark, body, ":(");
    }

    gtk_text_buffer_delete_mark (buffer, mark);
  }

  /**
   * Generates a GtkPixmap object from a given GMimePart that contains an image.
   * Used for displaying attached pictures inline.
   */
  GdkPixbuf* get_pixbuf_from_gmime_part (const GMimePart * part)
  {
    GdkPixbufLoader * l = gdk_pixbuf_loader_new ();

    // create the pixbuf
    size_t len (0);
    GdkPixbuf * pixbuf (0);
    GError * err (0);
    const char * content (g_mime_part_get_content (part, &len));
    gdk_pixbuf_loader_write (l, (const guchar*)content, len, &err);
    if (!err)
      pixbuf = gdk_pixbuf_loader_get_pixbuf (l);
    else {
      Log::add_err (err->message);
      g_clear_error (&err);
    }

    // cleanup
    gdk_pixbuf_loader_close (l, &err);
    if (err) {
      Log::add_err (err->message);
      g_clear_error (&err);
    }
    if (pixbuf)
      g_object_ref (G_OBJECT(pixbuf));
    g_object_unref (G_OBJECT(l));

    return pixbuf;
  }
}

char*
BodyPane :: get_fallback_charset ()
{
  //FIXME
  // some user interface means of setting a fallback charset
  return NULL;
}

namespace
{
  // as a last resort,
  // just tear out anything that's not UTF-8.
  char*
  scour_to_utf8 (const char * in,
                 int          in_len)
  {
    GString * str = g_string_new (NULL);
    const char *end;
    char * retval;

    while (g_utf8_validate (in, in_len, &end) == FALSE) {
      if (end > in)
        g_string_append_len (str, in, end-in);
      if (in_len != -1)
        in_len -= (1+end-in);
      in = end+1;
    }

    g_string_append_len (str, in, in_len);
    retval = str->str;
    g_string_free (str, FALSE);

    return retval;
  }
}

void
BodyPane :: append_part (GMimeObject * obj)
{
  bool is_done (false);

  // we only need leaf parts..
  if (!GMIME_IS_PART (obj))
    return;

  GMimePart * part = GMIME_PART (obj);
  const GMimeContentType * type = g_mime_object_get_content_type (GMIME_OBJECT (part));
  mime::YencInfo * yenc = static_cast<mime::YencInfo*>(g_object_get_data (G_OBJECT(obj), "yenc"));

  // if this part is yenc, handle it...
  // FIXME: was there some kind of recursion?  Why did we need Y_DECODER_INSTALLED?
  if (yenc!= NULL
    && part->content!=NULL
    && part->content->stream!=NULL
    && g_object_get_data(G_OBJECT(part),"Y_DECODER_INSTALLED")==NULL)
  {
    GMimeStream * stream = g_mime_stream_filter_new_with_stream (part->content->stream);
    GMimeFilter * yenc_filter = g_mime_filter_yenc_new (GMIME_FILTER_YENC_DIRECTION_DECODE);
    g_mime_stream_filter_add (GMIME_STREAM_FILTER(stream), yenc_filter);
    g_mime_data_wrapper_set_stream (part->content, stream);
    g_object_set_data (G_OBJECT(part), "Y_DECODER_INSTALLED", GINT_TO_POINTER(1));
    g_object_unref (yenc_filter);
    g_mime_stream_unref (stream);
  }

  // decide whether or not this part is a picture
  bool is_image (g_mime_content_type_is_type (type, "image", "*"));
  if (!is_image && g_mime_content_type_is_type(type, "application", "octet-stream")) {
    const char *type, *subtype;
    mime::guess_part_type_from_filename(g_mime_part_get_filename(part), &type, &subtype);
    is_image = type && !strcmp(type,"image");
  }

  // if it's a picture, draw it
  if (is_image) {
    GdkPixbuf * pixbuf = get_pixbuf_from_gmime_part (part);
    if (pixbuf != 0) {
      GtkTextIter iter;
      gtk_text_buffer_get_end_iter (_buffer, &iter);
      if (gtk_text_buffer_get_char_count (_buffer) > 0)
        gtk_text_buffer_insert (_buffer, &iter, "\n", -1);
      if (gtk_text_buffer_get_mark (_buffer, FIRST_PICTURE) == NULL)
        gtk_text_buffer_create_mark (_buffer, FIRST_PICTURE, &iter, TRUE);
      gtk_text_buffer_insert_pixbuf (_buffer, &iter, pixbuf);
      gtk_text_buffer_insert (_buffer, &iter, "\n", -1);
      g_object_unref (pixbuf);
      is_done = true;
    }
  }

  // or, if it's text, display it
  else if (g_mime_content_type_is_type (type, "text", "*"))
  {
    char * fallback_charset = get_fallback_charset ();
    char * str = mime_part_to_utf8 (part, fallback_charset);
    g_free (fallback_charset);

    const bool do_wrap (_prefs.get_flag ("wrap-article-body", false));
    if (str && *str && do_wrap) {
      std::string wrapped = _tm.fill (str);
      g_free (str);
      str = g_strdup (wrapped.c_str());
    }

    const bool do_mute (_prefs.get_flag ("mute-quoted-text", false));
    const bool do_smilies (_prefs.get_flag ("show-smilies-as-graphics", true));
    const bool do_markup (_prefs.get_flag ("show-text-markup", true));
    append_text_buffer_nolock (&_tm, _buffer, str, do_mute, do_smilies, do_markup);
    g_free (str);
    is_done = true;
  }

  // otherwise, bitch and moan.
  if (!is_done) {
    const char * filename = g_mime_part_get_filename (part);
    char * pch = (filename && *filename)
      ? g_strdup_printf (_("Attachment not shown: MIME type %s/%s; filename %s\n"), type->type, type->subtype, filename)
      : g_strdup_printf (_("Attachment not shown: MIME type %s/%s\n"), type->type, type->subtype);
    GtkTextIter iter;
    gtk_text_buffer_get_end_iter (_buffer, &iter);
    gtk_text_buffer_insert (_buffer, &iter, pch, -1);
    g_free (pch);
  }
}
void
BodyPane :: foreach_part_cb (GMimeObject* o, gpointer self)
{
  if (GMIME_IS_MULTIPART (o))
    g_mime_multipart_foreach (GMIME_MULTIPART (o), foreach_part_cb, self);
  else
    static_cast<BodyPane*>(self)->append_part (o);
}


/***
****  HEADERS
***/
namespace
{
  void add_bold_header_value (std::string   & s,
                              GMimeMessage  * message,
                              const char    * key,
                              const char    * fallback_charset)
  {
    const char * val (message ? g_mime_message_get_header (message, key) : "");
    char * utf8_val (header_to_utf8 (val, fallback_charset));
    char * e (0);
    if (strcmp (key, "From"))
      e = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", utf8_val);
    else {
      const StringView v = GNKSA :: get_short_author_name (utf8_val);
      e = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", v.to_string().c_str());
    }
    s += e;
    g_free (utf8_val);
    g_free (e);
  }

  size_t add_header_line (std::string& s, const char * key, const char * val, const char * fallback_charset)
  {
    char * e;
    e = g_markup_printf_escaped ("<span weight=\"bold\">%s:</span> ", key);
    s += e;
    g_free (e);
    char * utf8_val = header_to_utf8 (val, fallback_charset);
    e = g_markup_printf_escaped ("%s\n", utf8_val ? utf8_val : "");
    s += e;
    const size_t retval (g_utf8_strlen(key,-1) + g_utf8_strlen(utf8_val,-1) + 2);
    g_free (utf8_val);
    g_free (e);
    return retval;
  }

  size_t add_header_line (std::string& s, GMimeMessage * msg, const char * key, const char * fallback_charset)
  {
    const char * val (msg ? g_mime_message_get_header (msg, key) : "");
    return add_header_line (s, key, val, fallback_charset);
  }
}

void
BodyPane :: set_text_from_message (GMimeMessage * message)
{
  char * fallback_charset (get_fallback_charset());

  // set the headers...
  std::string s;
  size_t w (0);
  w = std::max (w, add_header_line (s, message, "Subject", fallback_charset));
  w = std::max (w, add_header_line (s, message, "From", fallback_charset));
  w = std::max (w, add_header_line (s, message, "Date", fallback_charset));
  s.resize (s.size()-1); // remove trailing linefeed
  gtk_label_set_markup (GTK_LABEL(_headers), s.c_str());

#if GTK_CHECK_VERSION(2,6,0)
  // ellipsize mode is useless w/o this in expander...
  gtk_label_set_width_chars (GTK_LABEL(_headers), (int)w);
#endif

  // set the x-face...
  GdkPixbuf * pixbuf (0);
  const char * pch = message ? g_mime_message_get_header (message, "X-Face") : NULL;
  if (pch != 0)
    pixbuf = pan_gdk_pixbuf_create_from_x_face (gtk_widget_get_colormap(_xface), _xface->window, pch);
  gtk_image_set_from_pixbuf (GTK_IMAGE(_xface), pixbuf);
  if (pixbuf)
    g_object_unref (pixbuf);

  // set the terse headers...
  s.clear ();
  add_bold_header_value (s, message, "Subject", fallback_charset);
  s += " from ";
  add_bold_header_value (s, message, "From", fallback_charset);
  s += " at ";
  add_bold_header_value (s, message, "Date", fallback_charset);
  gtk_label_set_markup (GTK_LABEL(_terse), s.c_str());
#if GTK_CHECK_VERSION(2,6,0)
  // ellipsize mode is useless w/o this in expander...
  gtk_label_set_width_chars (GTK_LABEL(_terse), (int)s.size());
#endif

  // clear the text buffer...
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (_buffer, &start, &end);
  gtk_text_buffer_delete (_buffer, &start, &end);
  if (gtk_text_buffer_get_mark (_buffer, FIRST_PICTURE) != NULL)
    gtk_text_buffer_delete_mark_by_name (_buffer, FIRST_PICTURE);

  // maybe add the headers
  const bool do_show_headers (_prefs.get_flag ("show-all-headers", false));
  if (message && do_show_headers) {
    char * headers = g_mime_message_get_headers (message);
    GtkTextIter end;
    gtk_text_buffer_get_end_iter (_buffer, &end);
    StringView line, v(headers);
    while (v.pop_token (line, '\n')) {
      char * pch = header_to_utf8 (line, fallback_charset);
      gtk_text_buffer_insert (_buffer, &end, pch, -1);
      gtk_text_buffer_insert (_buffer, &end, "\n", 1);
      g_free (pch);
    }
    gtk_text_buffer_insert (_buffer, &end, "\n", 1);
    g_free (headers);
  }

  // FIXME: need to set a mark here so that when user hits follow-up,
  // the all-headers don't get included in the followup

  // set the text buffer...
  if (message)
    g_mime_message_foreach_part (message, foreach_part_cb, this);

  // if there was a picture, scroll to it.
  // otherwise scroll to the top of the body.
  GtkTextMark * mark = gtk_text_buffer_get_mark (_buffer, FIRST_PICTURE);
  if (mark)
    gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW(_text), mark, 0.0, TRUE, 0.0, 0.0);
  else {
    GtkTextIter iter;
    gtk_text_buffer_get_start_iter  (_buffer, &iter);
    gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW(_text), &iter, 0.0, TRUE, 0.0, 0.0);
  }

  g_free (fallback_charset);
}

void
BodyPane :: refresh ()
{
  set_text_from_message (_message);
}

void
BodyPane :: set_article (const Article& a)
{
  _article = a;

  if (_message)
    g_object_unref (_message);
  _message = _cache.get_message (_article.parts);
  refresh ();

  _data.mark_read (_article);
}

void
BodyPane :: clear ()
{
  if (_message)
    g_object_unref (_message);
  _message = 0;
  refresh ();
}
  

/***
****
***/

gboolean
BodyPane :: expander_activated_idle (gpointer self_gpointer)
{
  BodyPane *  self (static_cast<BodyPane*>(self_gpointer));
  GtkExpander * ex (GTK_EXPANDER(self->_expander));
  const bool expanded = gtk_expander_get_expanded (ex);
  gtk_expander_set_label_widget (ex, expanded ? self->_verbose : self->_terse);
  return false;
}
void
BodyPane :: expander_activated_cb (GtkExpander*, gpointer self_gpointer)
{
  g_idle_add (expander_activated_idle, self_gpointer);
}

/***
****
***/

BodyPane :: BodyPane (Data& data, ArticleCache& cache, Prefs& prefs):
  _prefs (prefs),
  _data (data),
  _cache (cache),
  _message (0)
{
  GtkWidget * vbox = gtk_vbox_new (FALSE, PAD);
  gtk_container_set_resize_mode (GTK_CONTAINER(vbox), GTK_RESIZE_QUEUE);

  // about this expander... getting the ellipsis to work is a strange process.
  // once you turn ellipsize on, the expander tries to make its label as narrow
  // as it can and just have the three "..."s.  So gtk_label_set_width_chars
  // is used to force labels to want to be the right size... but then they
  // never ellipsize.  But, if we start with gtk_widget_set_size_request() to
  // tell the expander that it _wants_ to be very small, then it will still take
  // extra space given to it by its parent without asking for enough size to
  // fit the entire label.
  GtkWidget * w = _expander = gtk_expander_new (NULL);
  gtk_widget_set_size_request (w, 50, -1);
  g_signal_connect (w, "activate", G_CALLBACK(expander_activated_cb), this);
  gtk_box_pack_start (GTK_BOX(vbox), w, false, false, 0);

  _terse = gtk_label_new ("Expander");
  gtk_misc_set_alignment (GTK_MISC(_terse), 0.0f, 0.5f);
  gtk_label_set_use_markup (GTK_LABEL(_terse), true);
#if GTK_CHECK_VERSION(2,6,0)
  gtk_label_set_ellipsize (GTK_LABEL(_terse), PANGO_ELLIPSIZE_MIDDLE);
#endif
  gtk_widget_show (_terse);
  g_object_ref (_terse);

  GtkWidget * hbox = _verbose = gtk_hbox_new (FALSE, 0);
  w = _headers = gtk_label_new ("Headers");
  gtk_misc_set_alignment (GTK_MISC(w), 0.0f, 0.5f);
#if GTK_CHECK_VERSION(2,6,0)
  gtk_label_set_ellipsize (GTK_LABEL(w), PANGO_ELLIPSIZE_MIDDLE);
#endif
  gtk_label_set_use_markup (GTK_LABEL(w), true);
  gtk_box_pack_start (GTK_BOX(hbox), w, true, true, PAD_SMALL);
  w = _xface = gtk_image_new ();
  gtk_widget_set_size_request (w, 48, 48);
  gtk_box_pack_start (GTK_BOX(hbox), w, false, false, PAD_SMALL);
  gtk_widget_show_all (_verbose);
  g_object_ref (_verbose);

  // setup
  _text = gtk_text_view_new ();
  refresh_fonts ();
  gtk_widget_add_events (_text, GDK_POINTER_MOTION_MASK|GDK_POINTER_MOTION_HINT_MASK);
  gtk_container_set_border_width (GTK_CONTAINER(_text), PAD_SMALL);
  gtk_text_view_set_editable (GTK_TEXT_VIEW(_text), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW(_text), FALSE);
  GtkWidget * scroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(scroll), GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
                                  GTK_POLICY_AUTOMATIC,
                                  GTK_POLICY_AUTOMATIC);
  gtk_container_add (GTK_CONTAINER(scroll), _text);
  gtk_widget_show_all (vbox);
  gtk_box_pack_start (GTK_BOX(vbox), scroll, TRUE, TRUE, 0);

  // set up the buffer tags
  _buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW(_text));
  set_text_buffer_tags (_buffer, _prefs);

  // listen for user interaction
  g_signal_connect (_text, "motion_notify_event", G_CALLBACK(motion_notify_event), 0);
  g_signal_connect (_text, "button_press_event", G_CALLBACK(text_button_pressed), 0);
  g_signal_connect (_text, "key_press_event", G_CALLBACK(text_key_pressed), scroll);

  set_text_from_message (NULL);
  gtk_expander_set_expanded (GTK_EXPANDER(_expander), true);
  expander_activated_idle (this);

  _root = vbox;
  _prefs.add_listener (this);
  gtk_widget_show_all (_root);
}

BodyPane :: ~BodyPane ()
{
  _prefs.remove_listener (this);

  g_object_unref (_verbose);
  g_object_unref (_terse);

  if (_message)
    g_object_unref (_message);
}


namespace
{
  const int smooth_scrolling_speed (10);

  void sylpheed_textview_smooth_scroll_do (GtkAdjustment  * vadj,
                                           gfloat           old_value,
                                           gfloat           new_value,
                                           int              step)
  {
    const bool down (old_value < new_value);
    const int change_value = (int)(down ? new_value-old_value : old_value-new_value);
    for (int i=step; i<=change_value; i+=step)
      gtk_adjustment_set_value (vadj, old_value+(down?i:-i));
    gtk_adjustment_set_value (vadj, new_value);
  }
}

bool
BodyPane :: read_more_or_less (bool more)
{
  GtkWidget * parent = gtk_widget_get_parent (_text);
  GtkAdjustment * v = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW(parent));

  // figure out how far we scroll
  const int arbitrary_font_height_pixels_hack (18);
  const float inc (v->page_size - arbitrary_font_height_pixels_hack);
  const gfloat val (CLAMP (v->value + (more ? inc : -inc),
                           v->lower,
                           MAX(v->upper,v->page_size)-MIN(v->upper,v->page_size)));

  // if we can scroll, do so.
  bool handled (false);
  if (v->upper>=v->page_size && val!=v->value) {
    sylpheed_textview_smooth_scroll_do (v, v->value, val, smooth_scrolling_speed);
    handled = true;
  }

  return handled;
}

GMimeMessage*
BodyPane :: create_followup ()
{
  GMimeMessage * msg (0);

  if (_message)
  {
    msg = g_mime_message_new (false);

    ///
    ///  HEADERS
    ///

    // newsgroups
    const char * header = "Newsgroups";
    StringView v (g_mime_message_get_header (_message, header));
    g_mime_message_add_header (msg, header, v.str);

    // subject
    v = g_mime_message_get_subject (_message);
    std::string val (v.str, v.len);
    if (val.find ("Re:") != 0)
      val.insert (0, "Re: ");
    g_mime_message_set_subject (msg, val.c_str());

    // attribution line
    const char * cpch = g_mime_message_get_sender (_message);
    g_mime_message_add_header (msg, "X-Draft-Attribution-Author", cpch);
    cpch = g_mime_message_get_message_id (_message);
    g_mime_message_add_header (msg, "X-Draft-Attribution-Id", cpch);
    char * pch = g_mime_message_get_date_string (_message);
    g_mime_message_add_header (msg, "X-Draft-Attribution-Date", pch);
    g_free (pch);

    // references
    header = "References";
    v = g_mime_message_get_header (_message, header);
    val.assign (v.str, v.len);
std::cerr << LINE_ID << " val " << val << std::endl;
    if (!val.empty())
      val += ' ';
    val += "<";
    val += g_mime_message_get_message_id (_message);
    val += ">";
std::cerr << LINE_ID << " val " << val << std::endl;
    val = GNKSA :: trim_references (val);
std::cerr << LINE_ID << " val " << val << std::endl;
    g_mime_message_add_header (msg, header, val.c_str());

    ///
    ///  BODY
    ///

    GtkTextIter start, end;
    if (gtk_text_buffer_get_selection_bounds (_buffer, &start, &end))
    {
      // go with the user-selected region w/o modifications
      pch = gtk_text_buffer_get_text (_buffer, &start, &end, false);
    }
    else
    {
      // get the entire body...
      gtk_text_buffer_get_bounds (_buffer, &start, &end);
      pch = gtk_text_buffer_get_text (_buffer, &start, &end, false);

      // remove the sig...
      int sig_index (0);
      if (GNKSA::find_signature_delimiter (pch, sig_index) != GNKSA::SIG_NONE)
        pch[sig_index] = '\0';

      // trim...
      StringView v (pch);
      v.trim ();
      std::memmove (pch, v.str, v.len);
      pch[v.len] = '\0';
    }

    // quote the body
    std::string s;
    for (const char *c(pch); c && *c; ++c) {
      if (c==pch || c[-1]=='\n')
        s += (*c=='>' ? ">" : "> ");
      s += *c;
    }

    // set the clone's content object with our modified body
    GMimeStream * stream = g_mime_stream_mem_new ();
    g_mime_stream_write_string (stream, s.c_str());
    GMimeDataWrapper * wrapper = g_mime_data_wrapper_new_with_stream (stream, GMIME_PART_ENCODING_8BIT);
    GMimePart * part = g_mime_part_new ();
    GMimeContentType * type = g_mime_content_type_new_from_string ("text/plain; charset=UTF-8");
    g_mime_part_set_content_type (part, type);
    g_mime_part_set_content_object (part, wrapper);
    g_mime_part_set_encoding (part, GMIME_PART_ENCODING_8BIT);
    g_mime_message_set_mime_part (msg, GMIME_OBJECT(part));
    g_object_unref (wrapper);
    g_object_unref (part);
    g_object_unref (stream);
    g_free (pch);
//std::cerr << LINE_ID << " here is the modified clone\n" << g_mime_message_to_string(msg) << std::endl;
  }

  return msg;
}

/***
****
***/

void
BodyPane :: refresh_fonts ()
{
  const bool body_pane_font_enabled = _prefs.get_flag ("body-pane-font-enabled", false);
  const bool monospace_font_enabled = _prefs.get_flag ("monospace-font-enabled", false);

  if (!body_pane_font_enabled && !monospace_font_enabled)
    gtk_widget_modify_font (_text, 0);
  else {
    const std::string str (monospace_font_enabled
      ? _prefs.get_string ("monospace-font", "Monospace 10")
      : _prefs.get_string ("body-pane-font", "Sans 10"));
    PangoFontDescription * pfd (pango_font_description_from_string (str.c_str()));
    gtk_widget_modify_font (_text, pfd);
    pango_font_description_free (pfd);
  }
}

void
BodyPane :: on_prefs_flag_changed (const StringView& key, bool value)
{
  if ((key=="body-pane-font-enabled") || (key=="monospace-font-enabled"))
    refresh_fonts ();

  if ((key=="wrap-article-body") || (key=="mute-quoted-text") ||
      (key=="show-smilies-as-graphics") || (key=="show-all-headers") ||
      (key=="show-text-markup"))
    refresh ();
}

void
BodyPane :: on_prefs_string_changed (const StringView& key, const StringView& value)
{
  if ((key=="body-pane-font") || (key=="monospace-font"))
    refresh_fonts ();

}

void
BodyPane :: on_prefs_color_changed (const StringView& key, const GdkColor& color)
{
  if (key.strstr ("body-pane-color") != 0)
    refresh_colors ();
}

void
BodyPane :: refresh_colors ()
{
  set_text_buffer_tags (_buffer, _prefs);
  set_text_from_message (_message);
}
