/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <algorithm>
#include <cstdarg>
#include <gobject/gvaluecollector.h>
#include "pan-tree.h"

#define IS_SORTED(tree) \
  (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)

/***
****
***/
struct CellSetterImpl: public PanTreeStore::CellSetter
{
  private:

    PanTreeStore * tree;
    GtkTreeIter * iter;
    PanTreeStore::Node * node;

    GValue * get_gvalue (int column)
    {
      g_assert (0<= column && column < tree->n_columns);
      return &node->columns[column];
    }

  public:

    CellSetterImpl (PanTreeStore * t, GtkTreeIter * i)
    {
      g_assert (t);
      g_assert (i);
      g_assert (IS_PAN_TREE_STORE(t));
      g_assert (i->stamp == t->stamp);

      tree = t;
      iter = i;
      node = static_cast<PanTreeStore::Node*>(iter->user_data);
    }

    virtual ~CellSetterImpl () {}

  public:

    virtual bool set_cell_gvalue (int column, const GValue& gvalue)
    {
      g_return_val_if_fail (0<=column && column<tree->n_columns, false);

      // if value isn't of the right type, transform it.
      bool converted (false);
      GValue real_value = {0, };
      const GType in_type (G_VALUE_TYPE (&gvalue));
      const GType dest_type ((*tree->column_types)[column]);
      if (!g_type_is_a (in_type, dest_type)) {
        if (!(g_value_type_compatible (in_type, dest_type) &&
              g_value_type_compatible (dest_type, in_type))) {
          g_warning ("%s: Unable to convert from %s to %s\n",
                     G_STRLOC, g_type_name (in_type), g_type_name (dest_type));
          return false;
        }
        if (!g_value_transform (&gvalue, &real_value)) {
          g_warning ("%s: Unable to make conversion from %s to %s\n",
                     G_STRLOC, g_type_name (in_type), g_type_name (dest_type));
          g_value_unset (&real_value);
          return false;
        }
        converted = true;
      }

      // set the value
      if (!converted)
        g_value_copy (&gvalue, &node->columns[column]);
      else {
        g_value_copy (&real_value, &node->columns[column]);
        g_value_unset (&real_value);
      }
      return true;
    }

#define SET_VALUE(COLUMN,VALUE,TYPE,FUNC) \
    GValue * gval = get_gvalue (COLUMN); \
    if (G_VALUE_TYPE(gval) == TYPE) { \
      FUNC (gval, VALUE); \
      return true; \
    } else { \
      GValue tmp = { 0, }; \
      g_value_init (&tmp, TYPE); \
      FUNC (&tmp, VALUE); \
      const bool ok = set_cell_gvalue (column, tmp); \
      g_value_unset (&tmp); \
      return ok; \
    }
    bool set_cell_int (int column, int value) {
      SET_VALUE (column, value, G_TYPE_INT, g_value_set_int);
    }
    bool set_cell_uint (int column, unsigned int value) {
      SET_VALUE (column, value, G_TYPE_UINT, g_value_set_uint);
    }
    bool set_cell_string (int column, const char * value) {
      SET_VALUE (column, value, G_TYPE_STRING, g_value_set_string);
    }
    bool give_cell_string (int column, char * value) {
      SET_VALUE (column, value, G_TYPE_STRING, g_value_take_string);
    }
    bool set_cell_pointer (int column, void* value) {
      SET_VALUE (column, value, G_TYPE_POINTER, g_value_set_pointer);
    }
    bool set_cell_long (int column, long value) {
      SET_VALUE (column, value, G_TYPE_LONG, g_value_set_long);
    }
    bool set_cell_ulong (int column, unsigned long value) {
      SET_VALUE (column, value, G_TYPE_ULONG, g_value_set_ulong);
    }
    bool set_cell_char (int column, char value) {
      SET_VALUE (column, value, G_TYPE_CHAR, g_value_set_char);
    }
    bool set_cell_bool (int column, bool value) {
      SET_VALUE (column, value, G_TYPE_BOOLEAN, g_value_set_boolean);
    }
};

/***
****
***/

PanTreeStore :: Node*
PanTreeStore :: get_node (GtkTreeIter * iter)
{
  g_assert (iter);
  g_assert (iter->stamp == stamp);
  Node * n = static_cast<Node*>(iter->user_data);
  g_assert (n);
  return n;
}

const PanTreeStore :: Node*
PanTreeStore :: get_node (const GtkTreeIter * iter) const
{
  g_assert (iter);
  g_assert (iter->stamp == stamp);
  const Node * n = static_cast<const Node*>(iter->user_data);
  g_assert (n);
  return n;
}

namespace
{
  PanTreeStore::Node* get_node (PanTreeStore * tree, GtkTreeIter * iter)
  {
    g_assert (tree);
    g_assert (iter);
    g_assert (iter->stamp == tree->stamp);
    PanTreeStore::Node * n = static_cast<PanTreeStore::Node*>(iter->user_data);
    g_assert (n);
    return n;
  }
}

namespace
{
  void set_iter (GtkTreeIter   * iter,
                 int             stamp,
                 PanTreeStore::Node * node)
  {
    iter->stamp = stamp;
    iter->user_data = node;
    iter->user_data2 = 0;
    iter->user_data3 = 0;
  }

  void invalidate_iter (GtkTreeIter * iter)
  {
    set_iter (iter, 0, 0);
  }

  GtkTreeModelFlags get_flags (GtkTreeModel *tree_model)
  {
    return GTK_TREE_MODEL_ITERS_PERSIST;
  }

  gint get_n_columns (GtkTreeModel *tree)
  {
    return PAN_TREE_STORE(tree)->n_columns;
  }

  GType get_column_type (GtkTreeModel * tree,
                            gint           n)
  {
    return (*((PanTreeStore*)(tree))->column_types)[n];
  }

  gboolean get_iter (GtkTreeModel *tree_model,
                     GtkTreeIter  *iter,
                     GtkTreePath  *path)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (path, false);

    // find the article that corresponds to this path.
    PanTreeStore::Node * node = tree->root;
    const int depth = gtk_tree_path_get_depth (path);
    const int * indices = gtk_tree_path_get_indices (path);
    for (int i=0; i<depth; ++i) {
      const int index = *indices++;
      if (index<0 || index>=node->n_children)
        return false;
      node = node->children[index];
    }

    set_iter (iter, tree->stamp, node);
    return true;
  }

  GtkTreePath* get_path (GtkTreeModel *tree_model,
                         GtkTreeIter  *iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    GtkTreePath * path = gtk_tree_path_new ();
    PanTreeStore::Node * node = get_node (tree, iter);
    std::vector<int> indices;
    while (node && node!=tree->root) {
      indices.push_back (node->child_index);
      node = node->parent;
    }
    for (std::vector<int>::const_reverse_iterator it(indices.rbegin()), end(indices.rend()); it!=end; ++it)
      gtk_tree_path_append_index (path, *it);

    return path;
  }

  void get_value (GtkTreeModel * tree_model,
                  GtkTreeIter  * iter,
                  gint           column,
                  GValue       * dest_value)
  {
    g_assert (iter);
    g_assert (dest_value);
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_assert (tree);
    g_assert (iter->stamp == tree->stamp);
    g_assert (0<=column && column<tree->n_columns);

    PanTreeStore::Node * node = get_node (tree, iter);
    g_value_init (dest_value, (*tree->column_types)[column]);
    g_value_copy (&node->columns[column], dest_value);
  }

  gboolean iter_next (GtkTreeModel * model,
                      GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    PanTreeStore::Node * node = get_node (tree, iter);
    const int new_index = node->child_index + 1;
    if (new_index >= node->parent->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->parent->children[new_index]);
    return true;
  }

  gboolean iter_children (GtkTreeModel * model,
                          GtkTreeIter  * iter,
                          GtkTreeIter  * parent)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (!parent || (parent->stamp == tree->stamp), false);

    PanTreeStore::Node * node = parent ? get_node (tree, parent) : tree->root;
    if (node->children.empty()) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[0]);
    return true;
  }

  gint iter_n_children (GtkTreeModel * model,
                        GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, 0);
    g_return_val_if_fail (!iter || iter->stamp == tree->stamp, 0);

    const PanTreeStore::Node * node = iter ? get_node(tree, iter) : tree->root;
    return node->n_children;
  }

  gboolean iter_has_child (GtkTreeModel *model,
                           GtkTreeIter  *iter)
  {
    return iter_n_children (model, iter) != 0;
  }

  gboolean iter_nth_child (GtkTreeModel * model,
                           GtkTreeIter  * iter,
                           GtkTreeIter  * parent,
                           gint           n)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (!parent || parent->stamp == tree->stamp, false);

    PanTreeStore::Node * node = parent ? get_node (tree, parent) : tree->root;
    if (n<0 || n>=node->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[n]);
    return true;
  }
}

bool
PanTreeStore :: get_parent (GtkTreeIter * iter,
                            GtkTreeIter * child) const
{
  g_return_val_if_fail (child, false);
  g_return_val_if_fail (iter, false);
  g_return_val_if_fail (child->stamp == stamp, false);

  const PanTreeStore::Node * node = get_node (child);
  const bool has_parent = node->parent != root;
  if (has_parent)
    set_iter (iter, stamp, node->parent);
  else
    invalidate_iter (iter);
  return has_parent;
}

namespace
{
  gboolean iter_parent (GtkTreeModel * model,
                        GtkTreeIter  * iter,
                        GtkTreeIter  * child)
  {
    return PAN_TREE_STORE(model)->get_parent (iter, child);
  }
}

namespace
{
  gboolean get_sort_column_id (GtkTreeSortable  * sortable,
                               gint             * sort_column_id,
                               GtkSortType      * order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);

    if (sort_column_id)
       *sort_column_id = tree->sort_column_id;
    if (order)
       *order = tree->order;

    return (tree->sort_column_id != GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID)
        && (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID);
  }

  gboolean has_sort_func (GtkTreeSortable *sortable, int col)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);
    std::map<int,PanTreeStore::SortInfo>::const_iterator it =
      tree->sort_info->find (col);
    return (it!=tree->sort_info->end()) && (it->second.sort_func!=0);
  }

  struct SortTuple
  {
    int offset;
    PanTreeStore::Node * node;
    GtkTreeIter iter;
  };

  struct SortHelperData
  {
    PanTreeStore * store;
    GtkTreeModel * model;
    PanTreeStore::SortInfo& sort_info;
    GtkSortType order;
    SortHelperData (PanTreeStore           * tree_store,
                    PanTreeStore::SortInfo & s,
                    GtkSortType              o):
      store(tree_store), model(GTK_TREE_MODEL(tree_store)), sort_info(s), order(o) {}
  };

  int tuple_compare_func (gconstpointer a_gpointer,
                          gconstpointer b_gpointer,
                          gpointer user_data)
  {
    const SortTuple * a = (const SortTuple*) a_gpointer;
    const SortTuple * b = (const SortTuple*) b_gpointer;
    const SortHelperData * help = (const SortHelperData*) user_data;
    int val = (help->sort_info.sort_func) (help->model,
                                           const_cast<GtkTreeIter*>(&a->iter),
                                           const_cast<GtkTreeIter*>(&b->iter),
                                           help->sort_info.user_data);
    if (help->order == GTK_SORT_DESCENDING)
      val = -val;

    if (!val) // inplace sort
      val = get_node(help->store, const_cast<GtkTreeIter*>(&a->iter))->child_index - 
            get_node(help->store, const_cast<GtkTreeIter*>(&b->iter))->child_index;

    return val;
  }

  void pan_tree_store_sort_helper (PanTreeStore            * tree,
                                   PanTreeStore::SortInfo  & sort_info,
                                   PanTreeStore::Node      * parent,
                                   bool                      recurse)
  {
    const int n = parent->n_children;
    if (n < 2) // no need to sort [0...1] children
      return;

    // build a temporary array to sort
    SortTuple * sort_array = new SortTuple [n];
    for (int i=0; i<n; ++i) {
      SortTuple& tuple = sort_array[i];
      tuple.offset = i;
      tuple.node = parent->children[i];
      set_iter (&tuple.iter, tree->stamp, tuple.node);
    }

    // sort 'em...
    SortHelperData help (tree, sort_info, tree->order);
    g_qsort_with_data (sort_array, n,
                       sizeof(SortTuple),
                       tuple_compare_func,
                       &help);

    // update the child indices...
    for (int i=0; i<n; ++i) {
      PanTreeStore::Node * child = sort_array[i].node;
      child->child_index = i;
      parent->children[i] = child;
      g_assert (child->parent == parent);
    }

    // let the world know we've changed
    int * new_order = new int [n];
    bool differ = false;
    for (int i=0; i<n; ++i) {
      differ |= (i != sort_array[i].offset);
      new_order[i] = sort_array[i].offset;
    }

    // tell the world...
    GtkTreeIter tmp_iter, *parent_iter;
    GtkTreePath * path;
    if (parent == tree->root) {
      parent_iter = 0;
      path = gtk_tree_path_new ();
    } else {
      set_iter (&tmp_iter, tree->stamp, parent);
      parent_iter = &tmp_iter;
      path = gtk_tree_model_get_path (help.model, parent_iter);
    }
    gtk_tree_model_rows_reordered (help.model, path, parent_iter, new_order);
    gtk_tree_path_free (path);
    delete [] new_order;

    delete [] sort_array;

    for (int i=0; recurse && i<n; ++i) {
      PanTreeStore::Node * child = parent->children[i];
      pan_tree_store_sort_helper (tree, sort_info, child, recurse);
    }
  }

  void pan_tree_store_sort (PanTreeStore * tree)
  {
    g_assert (IS_PAN_TREE_STORE(tree));

    if (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)
    {
      g_assert (has_sort_func (GTK_TREE_SORTABLE(tree), tree->sort_column_id));
      PanTreeStore::SortInfo& info = (*tree->sort_info)[tree->sort_column_id];
      pan_tree_store_sort_helper (tree, info, tree->root, true);
    }
  }

  void pan_tree_store_sort_iter_changed (PanTreeStore      * tree,
                                         GtkTreeIter  * iter,
                                         gint           column)
  {
    // FIXME
  }


  void set_sort_column_id (GtkTreeSortable * sortable,
                           gint              sort_column_id,
                           GtkSortType       order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    // if no change, there's nothing to do...
    if (tree->sort_column_id==sort_column_id && tree->order==order)
      return;

    // sanity checks...
    if (sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID) {
      g_return_if_fail (tree->sort_info->count(sort_column_id) != 0);
      g_return_if_fail (tree->sort_info->find(sort_column_id)->second.sort_func != 0);
    }

    tree->sort_column_id = sort_column_id;
    tree->order = order;
    gtk_tree_sortable_sort_column_changed (sortable);
    pan_tree_store_sort (tree);
  }

  gboolean has_default_sort_func (GtkTreeSortable *sortable)
  {
    return has_sort_func (sortable, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID);
  }

  void set_sort_func (GtkTreeSortable        *sortable,
                      gint                    col,
                      GtkTreeIterCompareFunc  func,
                      gpointer                data,
                      GtkDestroyNotify        destroy)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    (*tree->sort_info)[col].assign (func, data, destroy);

    if (tree->sort_column_id == col)
      pan_tree_store_sort (tree);
  }

  void set_default_sort_func (GtkTreeSortable        * s,
                              GtkTreeIterCompareFunc   f,
                              gpointer                 p,
                              GtkDestroyNotify         d)
  {
    set_sort_func (s, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, f, p, d);
  }
}

namespace
{
  GObjectClass *parent_class (0);

  typedef std::vector<GValue> values_t;

  struct FreeNodeWalker: public PanTreeStore::WalkFunctor
  {
    PanTreeStore * store;
    const int n_columns;
    FreeNodeWalker (PanTreeStore *t): store(t), n_columns(t->n_columns) {}
    virtual ~FreeNodeWalker () {}
    virtual bool operator()(PanTreeStore*tree, GtkTreeIter*it, GtkTreePath *p) {
      PanTreeStore::Node * node = get_node (tree, it);
      g_assert (node->columns);
      GValue * v = node->columns;
      for (int i=0; i<n_columns; ++i, ++v)
        g_value_unset (v);
      delete [] node->columns;
      delete node;
      --tree->nodes_allocated;
      return true; // keep marching
    }
  };


  void
  pan_tree_finalize (GObject *object)
  {
    // erase the remaining nodes
    PanTreeStore * store = PAN_TREE_STORE(object);
    FreeNodeWalker walk (store);
    store->postfix_walk (NULL, walk);
    delete store->root;
    --store->nodes_allocated;
    g_assert (!store->nodes_allocated);
   
    delete store->column_types;
    delete store->sort_info;
    g_free (store->columns_template);

    (*parent_class->finalize) (object);
  }

  void
  pan_tree_class_init (PanTreeStoreClass *klass)
  {
    GObjectClass *object_class;

    parent_class = (GObjectClass*) g_type_class_peek_parent (klass);
    object_class = (GObjectClass*) klass;

    object_class->finalize = pan_tree_finalize;
  }

  void
  pan_tree_model_init (GtkTreeModelIface *iface)
  {
    iface->get_flags       = get_flags;
    iface->get_n_columns   = get_n_columns;
    iface->get_column_type = get_column_type;
    iface->get_iter        = get_iter;
    iface->get_path        = get_path;
    iface->get_value       = get_value;
    iface->iter_next       = iter_next;
    iface->iter_children   = iter_children;
    iface->iter_has_child  = iter_has_child;
    iface->iter_n_children = iter_n_children;
    iface->iter_nth_child  = iter_nth_child;
    iface->iter_parent     = iter_parent;
  }

  void
  pan_tree_sortable_init (GtkTreeSortableIface *iface)
  {
    iface->get_sort_column_id = get_sort_column_id;
    iface->set_sort_column_id = set_sort_column_id;
    iface->set_sort_func = set_sort_func;
    iface->set_default_sort_func = set_default_sort_func;
    iface->has_default_sort_func = has_default_sort_func;
  }

  void
  pan_tree_init (PanTreeStore * tree)
  {
    tree->stamp = g_random_int();
    tree->n_columns = 0;
    tree->root = new PanTreeStore::Node ();
    tree->nodes_allocated = 1;
    tree->column_types = new std::vector<GType>();
    tree->sort_info = new std::map<int,PanTreeStore::SortInfo>();
    tree->sort_column_id = GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID;
    tree->order = GTK_SORT_ASCENDING;
  }

  GValue*
  alloc_columns (const PanTreeStore * tree)
  {
    const int n_columns = tree->n_columns;
    GValue * columns = new GValue [n_columns];
    memcpy (columns, tree->columns_template, sizeof(GValue)*n_columns);
    return columns;
  }

  void
  pan_tree_alloc_nodes (PanTreeStore                     * tree,
                        size_t                             n_rows,
                        std::vector<PanTreeStore::Node*> & setme)
  {
    tree->nodes_allocated += n_rows;
    for (size_t i(0); i!=n_rows; ++i) {
      PanTreeStore::Node * node = new PanTreeStore::Node ();
      node->columns = alloc_columns (tree);
      setme.push_back (node);
    }
  }
}

/***
****
***/

GType
pan_tree_store_get_type (void)
{
  static GType pan_tree_type (0);
  if (pan_tree_type)
    return pan_tree_type;

  static const GTypeInfo pan_tree_info = {
    sizeof (PanTreeStoreClass),
    NULL, // base_init
    NULL, // base_finalize
    (GClassInitFunc) pan_tree_class_init,
    NULL, // class finalize
    NULL, // class_data
    sizeof (PanTreeStore),
    0, // n_preallocs
    (GInstanceInitFunc) pan_tree_init
  };

  pan_tree_type = g_type_register_static (
    G_TYPE_OBJECT, "PanTreeStore", &pan_tree_info, (GTypeFlags)0);
  static const GInterfaceInfo tree_model_info =
    { (GInterfaceInitFunc)pan_tree_model_init, 0, 0 };
  g_type_add_interface_static (
    pan_tree_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
  static const GInterfaceInfo sortable_info =
    { (GInterfaceInitFunc)pan_tree_sortable_init, 0, 0 };
  g_type_add_interface_static (
    pan_tree_type, GTK_TYPE_TREE_SORTABLE, &sortable_info);
  return pan_tree_type;
}

PanTreeStore *
pan_tree_store_new (int n_columns, ...)
{
  g_return_val_if_fail (n_columns>0, 0);
  PanTreeStore* tree = (PanTreeStore*) g_object_new (PAN_TREE_STORE_TYPE, NULL);

  va_list args;
  va_start (args, n_columns);
  for (int i=0; i<n_columns; ++i) {
    const GType type = va_arg (args, GType);
    tree->column_types->push_back (type);
  }
  va_end (args);
  tree->n_columns = tree->column_types->size();

  const int n = tree->n_columns;
  tree->columns_template = g_new0 (GValue, n);
  for (int i=0; i<n; ++i) {
    GValue& value = tree->columns_template[i];
    g_value_init (&value, (*tree->column_types)[i]);
  }

  return tree;
}

/***
****
***/

void
pan_tree_store_insert_n  (PanTreeStore               * tree,
                          GtkTreeIter                * iter,
                          GtkTreeIter                * parent_it,
                          PanTreeStore::RowSetter    & row_setter,
                          int                          n_rows,
                          int                          position)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (position >= 0);
  g_return_if_fail (n_rows > 0);

  // insert the rows
  PanTreeStore::Node * parent_node =
    parent_it ? get_node (tree, parent_it) : tree->root;
  std::vector<PanTreeStore::Node*> new_children;
  pan_tree_alloc_nodes (tree, n_rows, new_children);
  const int old_size = parent_node->n_children;
  position = std::min (position, old_size);
  parent_node->children.insert (parent_node->children.begin()+position,
                                new_children.begin(),
                                new_children.end());
  parent_node->n_children += n_rows;
  PanTreeStore::Node** it = &parent_node->children[position];
  for (int i=position, new_size=old_size+n_rows; i<new_size; ++i) {
    PanTreeStore::Node * node (*it++);
    node->child_index = i;
    node->parent = parent_node;
  }

  // set the return iter
  GtkTreeIter walk;
  set_iter (&walk, tree->stamp, parent_node->children[position]);
  if (iter)
    *iter = walk;

  // emit some signals
  GtkTreeModel * model = GTK_TREE_MODEL(tree);
  GtkTreePath * path = get_path (model, &walk);
  it = &parent_node->children[position];
  for (int i=0; i<n_rows; ++i) {
    PanTreeStore::Node * node (*it++);
    set_iter (&walk, tree->stamp, node);
    CellSetterImpl csi (tree, &walk);
    row_setter.set_row (i, csi);
    gtk_tree_model_row_inserted (model, path, &walk);
    gtk_tree_path_next (path);
  }
  if (!old_size && parent_node!=tree->root) {
    gtk_tree_path_up (path);
    gtk_tree_model_row_has_child_toggled (model, path, parent_it);
  }
  gtk_tree_path_free (path);
}

void
pan_tree_store_append_n  (PanTreeStore  * tree,
                          GtkTreeIter   * iter,
                          GtkTreeIter   * parent,
                          PanTreeStore::RowSetter     & row_setter,
                          int             n_rows)
{
  pan_tree_store_insert_n (tree, iter, parent, row_setter, n_rows, INT_MAX);
}

/***
****
***/

namespace 
{
  struct DoNothingRowSetter: public PanTreeStore::RowSetter {
    virtual ~DoNothingRowSetter() {}
    virtual void set_row (int n, PanTreeStore::CellSetter&)  {}
  };
}

void
pan_tree_store_insert (PanTreeStore * tree,
                       GtkTreeIter  * iter,
                       GtkTreeIter  * parent,
                       gint           pos)
{
  DoNothingRowSetter row_setter;
  pan_tree_store_insert_n (tree, iter, parent, row_setter, 1, pos);
}

void
pan_tree_store_insert_before (PanTreeStore * tree,
                              GtkTreeIter  * iter,
                              GtkTreeIter  * parent,
                              GtkTreeIter  * sibling)
{
  const int pos = sibling ? get_node(tree,sibling)->child_index : INT_MAX;
  pan_tree_store_insert (tree, iter, parent, pos);
}

void
pan_tree_store_insert_after (PanTreeStore * tree,
                             GtkTreeIter  * iter,
                             GtkTreeIter  * parent,
                             GtkTreeIter  * sibling)
{
  const int pos = sibling ? get_node(tree,sibling)->child_index+1 : 0;
  pan_tree_store_insert (tree, iter, parent, pos);
}

void pan_tree_store_prepend  (PanTreeStore * tree,
                              GtkTreeIter  * iter,
                              GtkTreeIter  * parent)
{
  pan_tree_store_insert (tree, iter, parent, 0);
}

void
pan_tree_store_append (PanTreeStore * tree,
                       GtkTreeIter * iter,
                       GtkTreeIter * parent)
{
  pan_tree_store_insert (tree, iter, parent, INT_MAX);
}

/***
****
***/

void
pan_tree_store_set (PanTreeStore * tree,
                    GtkTreeIter  *iter,
                    ...)
{
  va_list var_args;
  va_start (var_args, iter);
  pan_tree_store_set_valist (tree, iter, var_args);
  va_end (var_args);
}

void
pan_tree_store_set_valist (PanTreeStore * tree,
                           GtkTreeIter  * iter,
                           va_list        var_args)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (iter);
  g_return_if_fail (iter->stamp == tree->stamp);

  if (IS_SORTED(tree))
    g_return_if_fail (has_sort_func (GTK_TREE_SORTABLE(tree),
                      tree->sort_column_id));

  bool maybe_need_sort (false); //func != pan_tree_data_list_compare_func;
  bool emit_signal (false);
  int column = va_arg (var_args, gint);
  while (column != -1)
  {
    if (column >= tree->n_columns) {
      g_warning ("%s: Invalid column number %d added to iter (remember to end your list of columns with a -1)", G_STRLOC, column);
      break;
    }

    GValue value = { 0, };
    g_value_init (&value, (*tree->column_types)[column]);
    gchar * error = 0;
    G_VALUE_COLLECT (&value, var_args, 0, &error);
    if (error) {
      g_warning ("%s: %s", G_STRLOC, error);
      g_free (error);
      /* we purposely leak the value here, it might not be
       * in a sane state if an error condition occured */
      break;
    }

    CellSetterImpl csi (tree, iter);
    const bool did_set = csi.set_cell_gvalue (column, value);
    emit_signal |= did_set;
    g_value_unset (&value);

    maybe_need_sort |= column == tree->sort_column_id;

    column = va_arg (var_args, gint);
  }

  if (maybe_need_sort && IS_SORTED(tree))
    pan_tree_store_sort_iter_changed (tree, iter, tree->sort_column_id);

  if (emit_signal) {
    GtkTreePath * path = gtk_tree_model_get_path (GTK_TREE_MODEL(tree), iter);
    gtk_tree_model_row_changed (GTK_TREE_MODEL(tree), path, iter);
    gtk_tree_path_free (path);
  }
}

/***
****
***/

namespace
{
  struct NodeCompare
  {
    PanTreeStore::Node * root;
    NodeCompare (PanTreeStore::Node * r): root(r) {}

    int depth (const PanTreeStore::Node * node) const {
      int depth (0);
      while (node->parent != root) {
        ++depth;
        node = node->parent;
      }
      return depth;
    }

    // this sort function is very particular...
    // we go deepest first, so the first batches don't invalidate later ones.
    // we group by parent so to get groups of siblings for remove_siblings().
    // we sort by child_index so the first siblings removed in
    // remove_siblings() don't invalidate the tree path it uses.
    bool operator() (const PanTreeStore::Node* a,
                     const PanTreeStore::Node* b) const {
      // deepest first
      const int a_depth (depth (a));
      const int b_depth (depth (b));
      if (a_depth != b_depth)
        return b_depth < a_depth;
      // group by parent
      if (a->parent != b->parent)
        return a->parent < b->parent;
      // from last child to first
      return b->child_index < a->child_index;
    }
  };
}

void
PanTreeStore :: remove (GtkTreeIter * iters,
                        size_t        n_iters)
{
  g_return_if_fail (iters);
  g_return_if_fail (n_iters);

  Node ** nodes = new Node* [n_iters];
  for (size_t i=0; i<n_iters; ++i)
    nodes[i] = get_node (iters+i);

  NodeCompare compare (root);
  std::sort (nodes, nodes+n_iters, compare);

  Node * parent (0);
  Node ** children (0);
  for (Node **nit=nodes, **nend=nodes+n_iters; nit!=nend; ++nit) {
    if (parent && ((*nit)->parent != parent)) {
      remove_siblings (children, nit-children);
      parent = 0;
    }
    if (!parent) {
      parent = (*nit)->parent;
      g_assert (parent);
      children = nit;
    }
  }
  if (parent) {
    size_t n = (nodes + n_iters) - children;
    remove_siblings (children, n);
  }

  delete [] nodes;
}

void
PanTreeStore :: remove_siblings (Node ** siblings,
                                 size_t n_siblings)
{
  // entry assertions
  //g_assert (siblings);
  //g_assert (n_siblings);
  Node * parent = siblings[0]->parent;
  //for (Node **nit(siblings), **nend(nit+n_siblings); nit!=nend; ++nit) {
  //  g_assert ((*nit)->parent == parent); // all are siblings
  //  g_assert (nit==siblings || ((*nit)->child_index < nit[-1]->child_index));
  //}

  // get a path to the first sibilng.
  // we'll reuse it for all siblings.
  GtkTreeIter iter;
  set_iter (&iter, stamp, siblings[0]);
  GtkTreeModel * model (GTK_TREE_MODEL(this));
  GtkTreePath * path = gtk_tree_model_get_path (model, &iter);
  int * last_path_index = gtk_tree_path_get_indices (path);
  last_path_index += gtk_tree_path_get_depth(path)-1;

  // walk through all the siblings and unthread them.
  for (Node** nit(siblings), **nend(siblings+n_siblings); nit!=nend; ++nit)
  {
    Node * node (*nit);
    g_assert (node->parent);
    g_assert (node->parent == parent);

    // unthread the node
    const int pos = node->child_index;
    parent->children.erase (parent->children.begin() + pos);
    --parent->n_children;

    // let the world know the row was removed
    *last_path_index = node->child_index;
    gtk_tree_model_row_deleted (model, path);

    // free `node' and all its descendants
    FreeNodeWalker walk (this);
    GtkTreeIter iter;
    set_iter (&iter, stamp, *nit);
    postfix_walk (&iter, walk);
  }

  // renumber the surviving children
  if (parent->n_children) {
    int pos = 0;
    Node ** nit (&parent->children[pos]);
    for (int i=pos; i<parent->n_children; ++i, ++nit)
      (*nit)->child_index = i;
  }

  // if the parent's has-children state has changed, let the world know that too
  if (parent!=root && !parent->n_children) {
    GtkTreeIter pit;
    set_iter (&pit, stamp, parent);
    gtk_tree_path_up (path);
    gtk_tree_model_row_has_child_toggled (model, path, &pit);
  }

  gtk_tree_path_free (path);
}

/***
****
***/

#define get_cell_func(RETURN_TYPE, FUNC_NAME, VAL_FUNC_NAME) \
  RETURN_TYPE \
  PanTreeStore::FUNC_NAME (GtkTreeIter * iter, int column) const { \
    g_assert (iter); \
    g_assert (iter->stamp == stamp); \
    g_assert (0<=column && column<n_columns); \
    return VAL_FUNC_NAME (get_node(iter)->columns + column); \
  }

get_cell_func (int,            get_cell_int,     g_value_get_int)
get_cell_func (unsigned long,  get_cell_ulong,   g_value_get_ulong)
get_cell_func (gconstpointer,  get_cell_pointer, g_value_get_pointer)
get_cell_func (const char*,    get_cell_string,  g_value_get_string)

/***
****
***/

bool
PanTreeStore :: walk_helper (int           walk_mode,
                             Node        * node,
                             GtkTreePath * path,
                             WalkFunctor & walker)
{
  g_assert (node);
  g_assert (walk_mode==WALK_PREFIX || walk_mode==WALK_POSTFIX);

  bool more (true);

  GtkTreeIter iter;
  set_iter (&iter, stamp, node);

  if (more && node!=root && walk_mode==WALK_PREFIX)
    more = walker (this, &iter, path);

  const size_t n_children (node->n_children);
  if (more && n_children) {
    if (path)
      gtk_tree_path_append_index (path, 0);
    for (Node ** it(&node->children.front()),
              ** end(it+n_children); more && it!=end; ++it) {
      more = walk_helper (walk_mode, *it, path, walker);
      if (path)
        gtk_tree_path_next (path);
    }
    if (path)
      gtk_tree_path_up (path);
  }

  if (more && node!=root && walk_mode==WALK_POSTFIX)
    more = walker (this, &iter, path);

  return more;
}

void
PanTreeStore :: walk (int             walk_mode,
                      GtkTreeIter   * top,
                      WalkFunctor   & walker,
                      bool            need_path)
{
  PanTreeStore::Node * node = top ? get_node(top) : root;
  GtkTreePath * path = need_path ? gtk_tree_path_new() : 0;
  walk_helper (walk_mode, node, path, walker);
  gtk_tree_path_free (path);
}

void
PanTreeStore :: prefix_walk (GtkTreeIter   * top,
                             WalkFunctor   & walker,
                             bool            need_path)
{
  walk (WALK_PREFIX, top, walker, need_path);
}

void
PanTreeStore :: postfix_walk (GtkTreeIter   * top,
                              WalkFunctor   & walker,
                              bool            need_path)
{
  walk (WALK_POSTFIX, top, walker, need_path);
}

/**
***
**/

bool
PanTreeStore :: is_root (const GtkTreeIter* iter) const
{
  return iter && get_node(iter)->parent==root;
}
