/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <fstream>
#include <iostream>
#include <sstream>
#include <gmime/gmime.h>
#include <gtk/gtk.h>
#ifdef HAVE_GTKSPELL
extern "C" {
  #include <gtkspell/gtkspell.h>
}
#endif
#include <pan/general/debug.h>
#include <pan/general/file-util.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/general/log.h>
#include <pan/usenet-utils/gnksa.h>
#include <pan/usenet-utils/message-check.h>
#include <pan/usenet-utils/mime-utils.h>
#include <pan/data/data.h>
#include <pan/tasks/task-post.h>
#include "charset-picker.h"
#include "pad.h"
#include "hig.h"
#include "post-ui.h"
#include "post.ui.h"
#include "profiles-dialog.h"

using namespace pan;

/***
****  WRAP CODE
***/

namespace
{
  bool wrap_is_enabled (true);

  void on_wrap_toggled (GtkToggleAction* toggle, gpointer unused)
  {
    wrap_is_enabled = gtk_toggle_action_get_active (toggle);
  }

  void
  text_was_inserted_cb (GtkTextBuffer   * text_buffer,
                        GtkTextIter     * insert_pos,
                        char            * text,
                        int               text_len,
                        gpointer          tm_gpointer)
  {
    // dampen out a recursive event storm that would otherwise
    // happen from us changing the text buffer by rewrapping the
    // text here
    static bool dampen_feedback (false);

    if (dampen_feedback || !wrap_is_enabled)
      return;

    GtkTextIter line_begin_iter;
    const TextMassager& tm (*static_cast<TextMassager*>(tm_gpointer));
    const int line_number = gtk_text_iter_get_line (insert_pos);

    // make sure the line length isn't too wide
    gtk_text_buffer_get_iter_at_line (text_buffer, &line_begin_iter, line_number);
    int line_length = gtk_text_iter_get_chars_in_line(&line_begin_iter);
    if (line_length < tm.get_wrap_column())
      return;

    // we'll need to wrap all the way to the end of the paragraph to ensure
    // the wrapping works right -- so look for an empty line...
    GtkTextIter end_iter = line_begin_iter;
    for (;;) {
      if (!gtk_text_iter_forward_line (&end_iter))
        break;
      if (gtk_text_iter_get_chars_in_line (&end_iter) < 2)
        break;
    }

    // move backward to the end of the previous word
    gtk_text_iter_backward_word_start (&end_iter);
    gtk_text_iter_forward_word_end (&end_iter);

    // if the wrapped text is any different from the text already in
    // the buffer, then replace the buffer's text with the wrapped version.
    char * pch (gtk_text_buffer_get_text (text_buffer, &line_begin_iter, &end_iter, false));
    std::string before = pch;
    g_free (pch);
    std::string after (tm.fill (before));
    if (before == after)
      return;

    // okay, we need to rewrap.
    dampen_feedback = true;

    // remember where the insert pos was so that we can revalidate
    // the iterator that was passed in
    int insert_iterator_pos (gtk_text_iter_get_offset (insert_pos));

    // get char_offset so that we can update insert and selection_bound
    // marks after we make the change
    GtkTextMark * mark (gtk_text_buffer_get_mark (text_buffer, "insert"));
    GtkTextIter insert_iter;
    gtk_text_buffer_get_iter_at_mark (text_buffer, &insert_iter, mark);
    int char_offset (gtk_text_iter_get_offset (&insert_iter));

    // swap the non-wrapped for the wrapped text
    gtk_text_buffer_delete (text_buffer, &line_begin_iter, &end_iter);
    gtk_text_buffer_insert (text_buffer, &line_begin_iter, after.c_str(), -1);

    // update the insert and selection_bound marks to
    // where they were before the swap
    gtk_text_buffer_get_iter_at_offset (text_buffer, &insert_iter, char_offset);
    mark = gtk_text_buffer_get_mark (text_buffer, "insert");
    gtk_text_buffer_move_mark (text_buffer, mark, &insert_iter);
    mark = gtk_text_buffer_get_mark (text_buffer, "selection_bound");
    gtk_text_buffer_move_mark (text_buffer, mark, &insert_iter);

    // revalidate the insert_pos iterator
    gtk_text_buffer_get_iter_at_offset (text_buffer, insert_pos, insert_iterator_pos);

    dampen_feedback = false;
  }
}


/***
****  Menu and Toolbar
***/

namespace
{
  GtkWidget* get_focus (gpointer p) { return gtk_window_get_focus(GTK_WINDOW(static_cast<PostUI*>(p)->root())); }

  void do_cut      (GtkAction * w, gpointer p) { g_signal_emit_by_name (get_focus(p), "cut_clipboard"); }
  void do_copy     (GtkAction * w, gpointer p) { g_signal_emit_by_name (get_focus(p), "copy_clipboard"); }
  void do_paste    (GtkAction * w, gpointer p) { g_signal_emit_by_name (get_focus(p), "paste_clipboard"); }
  void do_rot13    (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->rot13_selection(); }
  void do_wrap     (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->wrap_body (); }
  void do_edit     (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->spawn_editor (); }
  void do_editors  (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->manage_editors (); }
  void do_profiles (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->manage_profiles (); }
  void do_post     (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->post_now (); }
  void do_save     (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->save_draft (); }
  void do_open     (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->open_draft (); }
  void do_close    (GtkAction * w, gpointer p) { static_cast<PostUI*>(p)->close_window (); }

  void editor_selected_cb (GtkRadioAction * action, GtkRadioAction * current, gpointer post_gpointer)
  {
    if (action == current)
    {
      const char * command = (const char*) g_object_get_data (G_OBJECT(action), "editor-command");
      static_cast<PostUI*>(post_gpointer)->set_editor_command (command);
    }
  }

  GtkActionEntry entries[] =
  {
    { "file-menu", NULL, _("_File") },
    { "edit-menu", NULL, _("_Edit") },
    { "profile-menu", NULL, _("_Profile") },
    { "editors-menu", NULL, _("Set Editor") },
    { "post-toolbar", NULL, "post" },
    { "post-article", GTK_STOCK_EXECUTE, _("_Post Article"), "<control>Return", _("Post this Article Now"), G_CALLBACK(do_post) },
    { "save-draft", GTK_STOCK_SAVE, _("_Save Draft"), "<control>s", _("Save as a Draft for Future Posting"), G_CALLBACK(do_save) },
    { "open-draft", GTK_STOCK_OPEN, _("_Open Draft..."), "<control>o", _("Open an Article Draft"), G_CALLBACK(do_open) },
    { "rewrap", GTK_STOCK_JUSTIFY_FILL, _("Wrap _Now"), NULL, _("Wrap the Article Body to 80 Columns"), G_CALLBACK(do_wrap) },
    { "close", GTK_STOCK_CLOSE, _("_Close"), NULL, _("Close Window"), G_CALLBACK(do_close) },
    { "cut", GTK_STOCK_CUT, _("Cu_t"), NULL, _("Cut Text"), G_CALLBACK(do_cut) },
    { "copy", GTK_STOCK_COPY, _("_Copy"), NULL, _("Copy Text"), G_CALLBACK(do_copy) },
    { "paste", GTK_STOCK_PASTE, _("_Paste"), NULL, _("Paste Text"), G_CALLBACK(do_paste) },
    { "rot13", GTK_STOCK_REFRESH, _("_Rot13"), NULL, _("Rot13 Selected Text"), G_CALLBACK(do_rot13) },
    { "run-editor", GTK_STOCK_JUMP_TO, _("Run _Editor"), "<control>e", NULL, G_CALLBACK(do_edit) },
    { "manage-editors", NULL, _("_Manage Editor List..."), NULL, NULL, G_CALLBACK(do_editors) },
    { "manage-profiles", NULL, _("Manage Posting Pr_ofiles..."), NULL, NULL, G_CALLBACK(do_profiles) }
  };

  GtkToggleActionEntry toggle_entries[] =
  {
    { "wrap", GTK_STOCK_JUSTIFY_FILL, _("_Wrap Text"), NULL, NULL, G_CALLBACK(on_wrap_toggled), true }
  };

  void add_widget (GtkUIManager* merge, GtkWidget* widget, gpointer vbox)
  {
    if (GTK_IS_TOOLBAR (widget)) {
      GtkWidget * handle_box = gtk_handle_box_new ();
      gtk_widget_show (handle_box);
      gtk_container_add (GTK_CONTAINER (handle_box), widget);
      g_signal_connect_swapped (widget, "destroy", G_CALLBACK (gtk_widget_destroy), handle_box);
      widget = handle_box;
    }
    gtk_box_pack_start (GTK_BOX(vbox), widget, false, false, 0);
  }
}

typedef Profiles::strings_t strings_t;

void
PostUI :: add_editor_list ()
{
  // remove the old editor buttons...
  typedef std::set<guint> unique_uints_t;
  static unique_uints_t editor_ui_ids;
  foreach_const (unique_uints_t, editor_ui_ids, it)
    gtk_ui_manager_remove_ui (_uim, *it);

  // add the new editor buttons...
  strings_t commands;
  _profiles.get_editors (commands);
  const std::string& editor_command (_profiles.get_active_editor ());
  GtkRadioAction * radio_group (0);
  for (size_t i=0; i<commands.size(); ++i)
  {
    const char * cmd (commands[i].c_str());
    char name[512], verb[512];
    g_snprintf (name, sizeof(name), "%s", cmd);
    g_snprintf (verb, sizeof(verb), "Edit With %s", cmd);


    GtkAction * o = (GtkAction*) g_object_new (GTK_TYPE_RADIO_ACTION,
       "name", verb,
       "label", name,
       "value", i,
       "group", radio_group, NULL);
    g_object_set_data_full (G_OBJECT(o), "editor-command", g_strdup(cmd), g_free);
    radio_group = GTK_RADIO_ACTION (o);
    const bool match (editor_command == cmd);
    if (match)
      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(o), true);
    g_signal_connect (o, "changed", G_CALLBACK(editor_selected_cb), this);
    gtk_action_group_add_action (_agroup, o);
    g_object_unref (o);


    const guint id (gtk_ui_manager_new_merge_id (_uim));
    gtk_ui_manager_add_ui (_uim, id,
                           "/post/edit-menu/editors-menu/editor-list",
                            name, verb,
                            GTK_UI_MANAGER_MENUITEM, false);
    editor_ui_ids.insert (id);
  }
}

void
PostUI :: add_actions (GtkWidget * box)
{
  _uim = gtk_ui_manager_new ();

  // read the file...
  char * filename = g_build_filename (file::get_pan_home().c_str(), "post.ui", NULL);
  GError * err (0);
  if (!gtk_ui_manager_add_ui_from_file (_uim, filename, &err)) {
    g_clear_error (&err);
    gtk_ui_manager_add_ui_from_string (_uim, fallback_post_ui, -1, &err);
  }
  if (err) {
    Log::add_err_va (_("Error loading post.ui: \"%s\""), err->message);
    g_clear_error (&err);
  }
  g_free (filename);

  g_signal_connect (_uim, "add_widget", G_CALLBACK(add_widget), box);

  // add the actions...
  _agroup = gtk_action_group_new ("post");
  gtk_action_group_add_actions (_agroup, entries, G_N_ELEMENTS(entries), this);
  gtk_action_group_add_toggle_actions (_agroup, toggle_entries, G_N_ELEMENTS(toggle_entries), this);
  gtk_ui_manager_insert_action_group (_uim, _agroup, 0);

  add_editor_list ();
}

void
PostUI :: set_editor_command (const StringView& command)
{
  _profiles.set_active_editor (command);
}

void
PostUI :: manage_profiles ()
{
  ProfilesDialog d (_profiles, GTK_WINDOW(_root));
  gtk_dialog_run (GTK_DIALOG(d.root()));
  gtk_widget_destroy (d.root());
  update_profile_combobox ();
  apply_profile ();
}

void
PostUI :: manage_editors ()
{
  GtkWidget * d = gtk_dialog_new_with_buttons (
    _("Manage Editor List"),
    GTK_WINDOW(_root),
    (GtkDialogFlags)(GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT),
    GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
    NULL);

  char b[512], lab[512];
  g_snprintf (b, sizeof(b), _("Editors"));
  g_snprintf (lab, sizeof(lab), "        <b>%s:</b>        ", b);
  GtkWidget * w = gtk_label_new (NULL);
  gtk_label_set_markup (GTK_LABEL(w), lab);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(d)->vbox), w, false, false, 0);
  gtk_widget_show (w);

  std::string s;
  strings_t commands;
  _profiles.get_editors (commands);
  foreach_const (strings_t, commands, it)
    s += *it + "\n";

  GtkWidget * frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);

  GtkTextBuffer * buf = gtk_text_buffer_new (NULL);
  gtk_text_buffer_set_text (buf, s.c_str(), s.size());
  w = gtk_text_view_new_with_buffer (buf);
  gtk_container_add (GTK_CONTAINER(frame), w);
  gtk_box_pack_start_defaults (GTK_BOX(GTK_DIALOG(d)->vbox), frame);
  gtk_container_set_border_width (GTK_CONTAINER(frame), PAD_BIG);
  gtk_widget_show_all (frame);

  gtk_dialog_run (GTK_DIALOG (d));

  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (buf, &start, &end);
  char * text (gtk_text_buffer_get_text (buf,  &start, &end, false));

  commands.clear ();
  StringView token, v (text);
  while (v.pop_token (token, '\n')) {
    token.trim ();
    if (!token.empty())
      commands.push_back (token.to_string());
  }

  _profiles.set_editors (commands);
  add_editor_list ();

  gtk_widget_destroy (d);
}

void
PostUI :: rot13_selection ()
{
  GtkTextIter start, end;
  if (gtk_text_buffer_get_selection_bounds (_body_buf, &start, &end))
  {
    char * str (gtk_text_buffer_get_text (_body_buf, &start, &end, false));
    _tm.rot13_inplace (str);
    gtk_text_buffer_delete (_body_buf, &start, &end);
    gtk_text_buffer_insert (_body_buf, &start, str, strlen(str));
    g_free (str);
  }
} 

void
PostUI :: wrap_body ()
{
  // get the current body
  GtkTextBuffer * buffer (_body_buf);
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (buffer, &start, &end);
  char * body = gtk_text_buffer_get_text (buffer, &start, &end, false);

  // wrap the body
  const std::string new_body (_tm.fill( body));

  // turn off our own wrapping while we fill the body pane */
  const bool b (wrap_is_enabled);
  wrap_is_enabled = false;
  gtk_text_buffer_set_text (_body_buf, new_body.c_str(), new_body.size());
  wrap_is_enabled = b;

  // cleanup
  g_free (body);
}

void
PostUI :: close_window ()
{
  gtk_widget_destroy (_root);
}

namespace
{
  void message_dialog_set_text (GtkMessageDialog * dialog,
                                const char * primary,
                                const char * secondary)
  {
#if GTK_CHECK_VERSION(2,6,0)
    gtk_message_dialog_set_markup (dialog, primary);
    gtk_message_dialog_format_secondary_text (dialog, "%s", secondary);
#else
    char * pch = g_strdup_printf ("<b>%s</b>\n\n%s", primary, secondary);
    gtk_message_dialog_set_markup (dialog, pch);
    g_free (pch);
#endif
  }
}

bool
PostUI :: check_message (const Quark& server, GMimeMessage * msg)
{
  MessageCheck :: unique_strings_t errors;
  MessageCheck :: Goodness goodness;

  quarks_t groups_this_server_has;
  _gs.server_get_groups (server, groups_this_server_has);
  MessageCheck :: message_check (msg, _hidden_headers["X-Draft-Attribution"], groups_this_server_has, errors, goodness);

  if (goodness.is_ok())
    return true;

  std::string s;
  foreach_const (MessageCheck::unique_strings_t, errors, it)
    s += *it + "\n";
  s.resize (s.size()-1); // eat trailing linefeed

  const GtkMessageType type (goodness.is_refuse() ? GTK_MESSAGE_ERROR : GTK_MESSAGE_WARNING);
  GtkWidget * d = gtk_message_dialog_new (GTK_WINDOW(_root),
                                          GTK_DIALOG_DESTROY_WITH_PARENT,
                                          type, GTK_BUTTONS_NONE, NULL);
  message_dialog_set_text (GTK_MESSAGE_DIALOG(d),
                           _("There were problems with this post."),
                           s.c_str());
  gtk_dialog_add_button (GTK_DIALOG(d), _("Go Back"), GTK_RESPONSE_CANCEL);
  if (goodness.is_warn())
    gtk_dialog_add_button (GTK_DIALOG(d), _("Post Anyway"), GTK_RESPONSE_APPLY);
  const int response = gtk_dialog_run (GTK_DIALOG(d));
  gtk_widget_destroy (d);
  return response == GTK_RESPONSE_APPLY;
}

namespace
{
  GtkWidget * new_go_online_button ()
  {
    GtkWidget * button = gtk_button_new ();
    GtkWidget * label = gtk_label_new_with_mnemonic (_("Go _Online"));
    GtkWidget * image = gtk_image_new_from_stock (GTK_STOCK_NETWORK, GTK_ICON_SIZE_BUTTON);
    GtkWidget * hbox = gtk_hbox_new (false, 2);
    GtkWidget * align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
    gtk_box_pack_start (GTK_BOX (hbox), image, false, false, 0);
    gtk_box_pack_end (GTK_BOX (hbox), label, false, false, 0);
    gtk_container_add (GTK_CONTAINER (align), hbox);
    gtk_container_add (GTK_CONTAINER (button), align);
    gtk_widget_show_all (button);
    return button;
  }
}

namespace
{
  gboolean pulse_me (gpointer progressbar)
  {
    gtk_progress_bar_pulse (GTK_PROGRESS_BAR (progressbar));
    return true;
  }
  void remove_progress_tag (gpointer tag)
  {
    g_source_remove (GPOINTER_TO_UINT(tag));
  }
}

void
PostUI :: on_progress_finished (Progress&, int status)
{
  _post_task->remove_listener (this);
  gtk_widget_destroy (_post_dialog);
  _post_dialog = 0;
  if (status == OK)
    close_window ();
  else {
    //if (status == FAIL)
    std::cerr << LINE_ID << " status " << status << std::endl;
  }
}

void
PostUI :: post_now ()
{
  GMimeMessage * message = new_message_from_ui (true);

  /**
  ***  Find a server that has all the groups we're posting to
  **/
  quarks_t groups;
  StringView token, v(g_mime_message_get_header (message, "Newsgroups"));
  while (v.pop_token (token, ',')) {
    token.trim ();
    groups.insert (token);
  }
  Quark server;
  if (!groups.empty()) {
    quarks_t servers;
    _gs.group_get_servers (*groups.begin(), servers);
    foreach_const (quarks_t, groups, it) {
      quarks_t s, diff;
      _gs.group_get_servers (*it, s);
      std::set_intersection (servers.begin(), servers.end(), s.begin(), s.end(), inserter(diff,diff.begin()));
      servers.swap (diff);
    }
    if (!servers.empty())
      server = *servers.begin ();
  }

  /**
  ***  Make sure the message is OK...
  **/
  if (!check_message (server, message)) {
    g_object_unref (G_OBJECT(message));
    return;
  }

  /**
  ***  Make sure we're online...
  **/
  if (!_queue.is_online())
  {
    GtkWidget * d = gtk_message_dialog_new (GTK_WINDOW(_root),
                                            GTK_DIALOG_DESTROY_WITH_PARENT,
                                            GTK_MESSAGE_QUESTION,
                                            GTK_BUTTONS_NONE, NULL);
    message_dialog_set_text (GTK_MESSAGE_DIALOG(d),
                             _("Pan is Offline."),
                             _("Go online to post the article?"));
    gtk_dialog_add_button (GTK_DIALOG(d), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
    gtk_dialog_add_action_widget (GTK_DIALOG(d), new_go_online_button(), GTK_RESPONSE_OK);
    const int response = gtk_dialog_run (GTK_DIALOG(d));
    gtk_widget_destroy (d);
    if (response !=  GTK_RESPONSE_OK) {
      g_object_unref (G_OBJECT(message));
      return;
    }
    _queue.set_online (true);
  }

  /**
  ***  Pop up a ``Posting'' Dialog...
  **/
  GtkWidget * d = gtk_dialog_new_with_buttons (_("Posting Article"),
                                               GTK_WINDOW(_root),
                                               GTK_DIALOG_DESTROY_WITH_PARENT,
                                               //GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                               NULL);
  char buf[512];
  g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("Posting..."));
  GtkWidget * w = GTK_WIDGET (g_object_new (GTK_TYPE_LABEL, "use-markup", TRUE, "label", buf, NULL));
  //GtkWidget * w = gtk_label_new (0);
  //gtk_label_set_markup (GTK_LABEL(w), buf);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(d)->vbox), w, false, false, PAD_SMALL);
  w = gtk_progress_bar_new ();
  gtk_progress_bar_set_pulse_step (GTK_PROGRESS_BAR(w), 0.05);
  const guint tag = g_timeout_add (100, pulse_me, w);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(d)->vbox), w, false, false, PAD_SMALL);
  g_object_set_data_full (G_OBJECT(d), "progressbar-timeout-tag", GUINT_TO_POINTER(tag), remove_progress_tag);
  _post_dialog = d;
  gtk_widget_show_all (d);
  _post_task = new TaskPost (server, message);
  _post_task->add_listener (this);
  
  _queue.add_task (_post_task);
  g_object_unref (message);
}

void
PostUI :: spawn_editor ()
{
  GtkTextBuffer * buf (_body_buf);
  bool ok (true);

  // open a new tmp file
  char * fname (0);
  FILE * fp (0);
  if (ok) {
    GError * err = NULL;
    const int fd (g_file_open_tmp ("pan_edit_XXXXXX", &fname, &err));
    if (fd != -1)
      fp = fdopen (fd, "w");
    else {
      Log::add_err (err && err->message ? err->message : _("Error opening temporary file"));
      if (err)
        g_clear_error (&err);
      ok = false;
    }
  }

  // get the current article body
  char * chars (0);
  size_t len;
  if (ok) {
    GtkTextIter start, end;
    gtk_text_buffer_get_bounds (buf, &start, &end);
    chars = gtk_text_buffer_get_text (buf, &start, &end, false);
    len = strlen (chars);
  }

  if (ok) {
    if (fwrite (chars, sizeof(char), len, fp) != len) {
      ok = false;
      Log::add_err_va (_("Error writing article to temporary file: %s"), g_strerror(errno));
    }
  }

  if (fp != NULL) {
    fclose (fp);
    fp = NULL;
  }

  g_free (chars);

  // parse the command line
  int argc (0);
  char ** argv (0);
  if (ok) {
    const std::string& editor (_profiles.get_active_editor());
    GError * err (0);
    g_shell_parse_argv (editor.c_str(), &argc, &argv, &err);
    if (err != NULL) {
      Log::add_err_va (_("Error parsing \"external editor\" command line: %s (Command was: %s)"), err->message, editor.c_str());
      g_clear_error (&err);
      ok = false;
    }
  }

  // put temp file's name into the substitution
  bool filename_added (false);
  for (int i=0; i<argc; ++i) {
    char * token (argv[i]);
    char * sub (strstr (token, "%t"));
    if (sub) {
      GString * gstr  = g_string_new (0);
      g_string_append_len (gstr, token, sub-token);
      g_string_append (gstr, fname);
      g_string_append (gstr, sub+2);
      g_free (token);
      argv[i] = g_string_free (gstr, false);
      filename_added = true;
    }
  }

  // no substitution field -- add the filename at the end
  if (!filename_added) {
    char ** v = g_new (char*, argc+2);
    for (int i=0; i<argc; ++i)
      v[i] = argv[i];
    v[argc++] = g_strdup (fname);
    v[argc] = NULL;
    g_free (argv);
    argv = v;
  }

  // spawn off the external editor
  if (ok) {
    GError * err (0);
    g_spawn_sync (0, argv, 0, G_SPAWN_SEARCH_PATH, 0, 0, 0, 0, 0, &err);
    if (err != NULL) {
      Log::add_err_va (_("Error starting external editor: %s"), err->message);
      g_clear_error (&err);
      ok = false;
    }
  }

  // read the file contents back in
  if (ok) {
    GError * err (0);
    char * body (0);
    gsize body_len (0);
    g_file_get_contents (fname, &body, &body_len, &err);
    if (err != NULL) {
      Log::add_err_va (_("Error reading file \"%s\": %s"), err->message, g_strerror(errno));
      g_clear_error (&err);
      ok = false;
    } else {
      GtkTextIter start, end;
      gtk_text_buffer_get_bounds (buf, &start, &end);
      gtk_text_buffer_delete (buf, &start, &end);
      gtk_text_buffer_insert (buf, &start, body, body_len);
    }
    g_free (body);
  }

  // cleanup
  ::remove (fname);
  g_free (fname);
  g_strfreev (argv);
}

namespace
{
  std::string& get_draft_filename ()
  {
    static std::string fname;

    if (fname.empty())
    {
      fname = file::get_pan_home ();
      char * pch = g_build_filename (fname.c_str(), "article-drafts", NULL);
      file :: ensure_dir_exists (pch);
      fname = pch;
      g_free (pch);
    }

    return fname;
  }
}

void
PostUI :: open_draft ()
{
  std::string& draft_filename (get_draft_filename ());

  GtkWidget * d = gtk_file_chooser_dialog_new (_("Open Draft Article"),
                                               GTK_WINDOW(_root),
                                               GTK_FILE_CHOOSER_ACTION_OPEN,
                                               GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                               GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                               NULL);
  if (!gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(d), draft_filename.c_str()))
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(d), draft_filename.c_str());
  if (gtk_dialog_run(GTK_DIALOG(d)) == GTK_RESPONSE_ACCEPT)
  {
    char * pch = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(d));
    draft_filename = pch;
    g_free (pch);

    if (g_file_get_contents (draft_filename.c_str(), &pch, NULL, NULL))
    {
      GMimeStream * stream = g_mime_stream_mem_new_with_buffer (pch, strlen(pch));
      GMimeParser * parser = g_mime_parser_new_with_stream (stream);
      GMimeMessage * message = g_mime_parser_construct_message (parser);
      if (message) {
        set_message (message);
        g_object_unref (G_OBJECT(message));
      }
      g_object_unref (G_OBJECT(parser));
      g_object_unref (G_OBJECT(stream));
      g_free (pch);
    }

  }
  gtk_widget_destroy (d);
}

namespace
{
  std::string get_user_agent ()
  {
    std::string s = PACKAGE_STRING;
    s += " (";
    s += VERSION_TITLE;
    s += ')';
    return s;
  }

  bool extra_header_is_editable (const StringView& name,
                                 const StringView& value)
  {
    const bool keep = !name.empty() // make sure header exists
        && (name != "Subject") // not an extra header
        && (name != "Newsgroups")
        && (name != "From")
        && (name != "Xref") // not editable
        && (name != "Message-Id")
        && (name != "References")
        && ((name != "User-Agent") || (value != get_user_agent()))
        && (name.strstr ("Content-") != name.str)
        && (name.strstr ("X-Draft-") != name.str);
    return keep;
  }

#if !GTK_CHECK_VERSION(2,6,0)
  char* gtk_combo_box_get_active_text (GtkComboBox * combo_box)
  {
    char * text (0);
    GtkTreeIter  iter;
    if (gtk_combo_box_get_active_iter (combo_box, &iter)) {
      GtkTreeModel * model = gtk_combo_box_get_model (combo_box);
      gtk_tree_model_get (model, &iter, 0, &text, -1);
    }
    return text;
  }
#endif
}

GMimeMessage*
PostUI :: new_message_from_ui (bool strip_draft_headers)
{
  GMimeMessage * msg (g_mime_message_new (false));

  // headers from the ui: From
  Profile profile;
  char * pch = gtk_combo_box_get_active_text (GTK_COMBO_BOX(_from_combo));
  std::string s;
  if (pch) {
    _profiles.get_profile (pch, profile);
    profile.get_from_header (s);
    g_free (pch);
  }
  g_mime_message_add_header (msg, "From", s.c_str());

  // headers from the ui: Subject
  const char * cpch (gtk_entry_get_text (GTK_ENTRY(_subject_entry)));
  g_mime_message_set_subject (msg, cpch);

  // headers from the ui: Newsgroups
  cpch = gtk_entry_get_text (GTK_ENTRY(_groups_entry));
  g_mime_message_set_header (msg, "Newsgroups", cpch);

  // add the 'hidden headers'
  foreach_const (str2str_t, _hidden_headers, it)
    if (!strip_draft_headers || (it->first.find ("X-Draft-")!=0))
      g_mime_message_set_header (msg, it->first.c_str(), it->second.c_str());

  // build headers from the 'more headers' entry field
  std::map<std::string,std::string> headers;
  GtkTextBuffer * buf (_headers_buf);
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (buf, &start, &end);
  pch = gtk_text_buffer_get_text (buf, &start, &end, false);
  StringView key, val, v(pch);
  v.trim ();
  while (v.pop_token (val, '\n') && val.pop_token(key,':')) {
    key.trim ();
    val.eat_chars (1);
    val.trim ();
    std::string key_str (key.to_string());
    if (extra_header_is_editable (key, val))
      g_mime_message_set_header (msg, key.to_string().c_str(),
                                      val.to_string().c_str());
  }
  g_free (pch);

  // User-Agent
  if (!g_mime_message_get_header (msg, "User-Agent"))
    g_mime_message_set_header (msg, "User-Agent", get_user_agent().c_str());

  // body
  buf = _body_buf;
  gtk_text_buffer_get_bounds (buf, &start, &end);
  pch = gtk_text_buffer_get_text (buf, &start, &end, false);
  pch = g_strstrip (pch);
  GMimeStream * stream = g_mime_stream_mem_new ();
  g_mime_stream_write_string (stream, pch);
  GMimeDataWrapper * wrapper = g_mime_data_wrapper_new_with_stream (stream, GMIME_PART_ENCODING_8BIT);
  GMimePart * part = g_mime_part_new ();
  GMimeContentType * type = g_mime_content_type_new_from_string ("text/plain; charset=UTF-8");
  g_mime_part_set_content_type (part, type);
  g_mime_part_set_content_object (part, wrapper);
  g_mime_part_set_encoding (part, GMIME_PART_ENCODING_8BIT);
  g_mime_message_set_mime_part (msg, GMIME_OBJECT(part));
  g_object_unref (wrapper);
  g_object_unref (part);
  g_object_unref (stream);
  g_free (pch);

  return msg;
}

void
PostUI :: save_draft ()
{
  std::string& draft_filename (get_draft_filename ());

  GtkWidget * d = gtk_file_chooser_dialog_new (
    _("Save Draft Article"),
    GTK_WINDOW(_root),
    GTK_FILE_CHOOSER_ACTION_SAVE,
    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
    NULL);

  if (!gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(d), draft_filename.c_str()))
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(d), draft_filename.c_str());

  if (gtk_dialog_run(GTK_DIALOG(d)) == GTK_RESPONSE_ACCEPT)
  {
    GMimeMessage * msg = new_message_from_ui (false);
    char * filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(d));

    errno = 0;
    std::ofstream o (filename);
    char * pch = g_mime_message_to_string (msg);
    o << pch;
    o.close ();

    if (o.fail()) {
      GtkWidget * e = gtk_message_dialog_new (
        GTK_WINDOW(d),
        GTK_DIALOG_DESTROY_WITH_PARENT,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_CLOSE,
        _("Unable to save \"%s\" %s"), filename, file::pan_strerror(errno));
      gtk_dialog_run (GTK_DIALOG(e));
      gtk_widget_destroy (e);
    }

    g_free (pch);
    g_free (filename);
    g_object_unref (msg);
  }

  gtk_widget_destroy (d);
}

namespace
{
  GtkWidget* create_body_widget (GtkTextBuffer*& buf, GtkWidget*& view, const Prefs &prefs)
  {
    view = gtk_text_view_new ();

#ifdef HAVE_GTKSPELL
    GError * err (0);
    const char * locale = NULL;
    gtkspell_new_attach (GTK_TEXT_VIEW(view), locale, &err);
    if (err) {
      Log::add_err_va (_("Error setting spellchecker: %s"), err->message);
      g_clear_error (&err);
    }
#endif

    buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(view));

    // force a monospace font, and size it to 80 cols x 30 rows
	const std::string str (prefs.get_string ("monospace-font", "Monospace 10"));
	PangoFontDescription *pfd (pango_font_description_from_string (str.c_str()));
    PangoContext * context = gtk_widget_create_pango_context (view);
    const int column_width (80);
    std::string line (column_width, 'A');
    pango_context_set_font_description (context, pfd);
    PangoLayout * layout = pango_layout_new (context);
    pango_layout_set_text (layout, line.c_str(), line.size());
    PangoRectangle r;
    pango_layout_get_extents (layout, &r, 0);
    gtk_widget_set_size_request (view, PANGO_PIXELS(r.width),
                                       PANGO_PIXELS(r.height*30));
    gtk_widget_modify_font (view, pfd);
    
    gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(view), GTK_WRAP_NONE);
    gtk_text_view_set_editable (GTK_TEXT_VIEW(view), true);
    GtkWidget * scrolled_window = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(scrolled_window),
                                         GTK_SHADOW_IN);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC);
    gtk_container_add (GTK_CONTAINER (scrolled_window), view);

    // cleanup
    g_object_unref (G_OBJECT(layout));
    g_object_unref (G_OBJECT(context));
    pango_font_description_free (pfd);

    return scrolled_window;
  }
}

void
PostUI :: update_profile_combobox ()
{
  // get the old selection
  GtkComboBox * combo (GTK_COMBO_BOX (_from_combo));
  char * active_text = gtk_combo_box_get_active_text (combo);

  // tear out the old entries
  GtkTreeModel * model (gtk_combo_box_get_model (combo));
  for (int i(0), qty(gtk_tree_model_iter_n_children(model,0)); i<qty; ++i)
    gtk_combo_box_remove_text (combo, 0);

  // add the new entries
  typedef std::set<std::string> names_t;
  const names_t profile_names (_profiles.get_profile_names ());
  int index (0);
  int sel_index (0);
  foreach_const (names_t, profile_names, it) {
    gtk_combo_box_append_text (combo, it->c_str());
    if (active_text && (*it == active_text))
      sel_index = index;
    ++index;
  }
 
  // ensure _something_ is selected...
  gtk_combo_box_set_active (combo, sel_index);

  // cleanup
  g_free (active_text);
}

namespace
{
  std::string load_signature (const Profile& profile)
  {
    std::string s;
    if (profile.use_sigfile && !profile.signature_file.empty()) {
      char * pch (0);
      if (g_file_get_contents (profile.signature_file.c_str(), &pch, 0, 0))
        s = pch;
      else
        s = profile.signature_file;
      g_free (pch);
    }
    return s;
  }
}

void
PostUI :: apply_profile ()
{
  apply_profile_to_body ();
  apply_profile_to_headers ();
}

namespace
{
  bool do_attribution_substitutions (const StringView & mid,
                                     const StringView & date,
                                     const StringView & from,
                                     std::string& attrib)
  {
    std::string::size_type pos;

    if (mid.empty() && date.empty() && from.empty()) // not a follow-up; attribution not needed
      return false;

    while (!mid.empty() && (((pos = attrib.find ("%i"))) != std::string::npos))
      attrib.replace (pos, 2, mid);

    while (!date.empty() && (((pos = attrib.find ("%d"))) != std::string::npos))
      attrib.replace (pos, 2, date);

    while (!from.empty() && (((pos = attrib.find ("%a"))) != std::string::npos))
      attrib.replace (pos, 2, from);

    StringView brief = GNKSA :: get_short_author_name (from);
    while (!brief.empty() && (((pos = attrib.find ("%n"))) != std::string::npos))
      attrib.replace (pos, 2, brief.str, brief.len);

    return true;
  }
}

void
PostUI :: apply_profile_to_body ()
{
  // get the selected profile
  Profile profile;
  char * pch = gtk_combo_box_get_active_text (GTK_COMBO_BOX(_from_combo));
  if (pch) {
    _profiles.get_profile (pch, profile);
    g_free (pch);
  }
  std::string attribution = profile.attribution;
  if (do_attribution_substitutions (_hidden_headers["X-Draft-Attribution-Id"],
                                    _hidden_headers["X-Draft-Attribution-Date"],
                                    _hidden_headers["X-Draft-Attribution-Author"],
                                    attribution))
    attribution = _tm.fill (attribution);
  else
    attribution.clear ();

  // get current the body
  GtkTextBuffer * buf (_body_buf);
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (buf, &start, &end);
  pch = gtk_text_buffer_get_text (buf, &start, &end, false);
  std::string body;
  if (pch)
    body = pch;
  g_free (pch);

  // replace the attribution
  const std::string old_attribution (_hidden_headers["X-Draft-Attribution"]);
  if (!attribution.empty()) {
    std::string::size_type pos = body.find (old_attribution);
    if (!old_attribution.empty() && (pos != std::string::npos))
      body.replace (pos, old_attribution.size(), attribution);
    else if (!attribution.empty()) {
      body = attribution + "\n\n" + body;
    }
    _hidden_headers["X-Draft-Attribution"] = attribution;
  }

  // remove the last signature
  int index;
  if (!GNKSA :: find_signature_delimiter (body, index))
    index = body.size();
  StringView v (body.c_str(), index);
  v.trim ();
  body.assign (v.str, v.len);
  if (!body.empty())
    body += "\n\n";
  const int insert_pos = body.size();

  // insert the new signature
  std::string sig = load_signature (profile);
  if (!sig.empty()) {
    int ignored;
    if (GNKSA::find_signature_delimiter (sig, ignored) == GNKSA::SIG_NONE)
      body += "\n\n-- \n";
    body += sig;
  }

  gtk_text_buffer_set_text (buf, body.c_str(), body.size()); // FIXME: wrap

  // set & scroll-to the insert point
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_offset (buf, &iter, insert_pos);
  gtk_text_buffer_move_mark_by_name (buf, "insert", &iter);
  gtk_text_buffer_move_mark_by_name (buf, "selection_bound", &iter);
  gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW(_body_view),
                                gtk_text_buffer_get_mark(buf, "insert"),
                                0.0, false, 0.0, 0.0);
}

void
PostUI :: apply_profile_to_headers ()
{
  // get the selected profile
  Profile profile;
  char * pch = gtk_combo_box_get_active_text (GTK_COMBO_BOX(_from_combo));
  if (pch) {
    _profiles.get_profile (pch, profile);
    g_free (pch);
  }

  // get the current `extra headers'
  GtkTextBuffer * buf (_headers_buf);
  Profile::headers_t headers;
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (buf, &start, &end);
  char * text (gtk_text_buffer_get_text (buf,  &start, &end, false));
  StringView key, val, v(text);
  while (v.pop_token (val, '\n') && val.pop_token(key,':')) {
    key.trim ();
    val.eat_chars (1);
    val.trim ();
    headers[key.to_string()].assign (val.str, val.len);
  }
  g_free (text);

  // we want to propagate changes from the profile manager to here,
  // so override the text field with the profile's values...
  foreach_const (Profile::headers_t, profile.headers, it)
    headers[it->first] = it->second;

  // rewrite the extra headers pane
  std::string s;
  foreach_const (Profile::headers_t, headers, it)
    s += it->first + ": " + it->second + "\n";
  gtk_text_buffer_set_text (buf, s.c_str(), s.size());
}

namespace
{
  void delete_post_ui (gpointer p)
  {
    delete static_cast<PostUI*>(p);
  }

  void on_from_combo_changed (GtkComboBox *widget, gpointer user_data)
  {
    static_cast<PostUI*>(user_data)->apply_profile ();
  }

  typedef std::map <std::string, std::string> str2str_t;

  struct SetMessageForeachHeaderData
  {
    str2str_t hidden_headers;
    std::string visible_headers;
  };

  void set_message_foreach_header_func (const char * name, const char * value, gpointer data_gpointer)
  {
    struct SetMessageForeachHeaderData * data (static_cast<SetMessageForeachHeaderData*>(data_gpointer));
    if (!extra_header_is_editable (name, value))
      data->hidden_headers[name] = value;
    else {
      std::string& str (data->visible_headers);
      str += std::string(name) + ": " + value + "\n";
    }
  }
}

void
PostUI :: set_message (GMimeMessage * message)
{
  // update our message header
  if (message)
    g_object_ref (G_OBJECT(message));
  if (_message)
    g_object_unref (G_OBJECT(_message));
  _message = message;

  // update subject, newsgroups fields
  const char * pch = g_mime_message_get_subject (message);
  gtk_entry_set_text (GTK_ENTRY(_subject_entry), pch ? pch : "");
  pch = g_mime_message_get_header (message, "Newsgroups");
  gtk_entry_set_text (GTK_ENTRY(_groups_entry), pch ? pch : "");

  // update 'other headers'
  SetMessageForeachHeaderData data;
  if (message->mime_part && g_mime_header_has_raw (message->mime_part->headers))
    g_mime_header_foreach (message->mime_part->headers, set_message_foreach_header_func, &data);
  g_mime_header_foreach (GMIME_OBJECT(message)->headers, set_message_foreach_header_func, &data);
  gtk_text_buffer_set_text (_headers_buf, data.visible_headers.c_str(), -1);
  _hidden_headers = data.hidden_headers;

  // update body
  int ignored;
  char * tmp = g_mime_message_get_body (message, true, &ignored);
  if (tmp) {
    std::string s = tmp;
    s += "\n\n";
    gtk_text_buffer_set_text (_body_buf, s.c_str(), s.size());
  }
  g_free (tmp);

  // apply the profiles
  update_profile_combobox ();
  apply_profile ();
}

PostUI :: ~PostUI ()
{
  g_object_unref (G_OBJECT(_message));
}

PostUI :: PostUI (Queue        & queue,
                  GroupServer  & gs,
                  Profiles     & profiles,
                  GMimeMessage * message,
                  const Prefs  & prefs,
                  GtkWindow    * parent) :
  _queue (queue),
  _gs (gs),
  _profiles (profiles),
  _root (0),
  _from_combo (0),
  _subject_entry (0),
  _groups_entry (0),
  _body_view (0),
  _body_buf (0),
  _message (0)
{
  GtkTooltips * tips = gtk_tooltips_new ();

  g_return_if_fail (message != 0);

  // create the window
  _root = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_role (GTK_WINDOW(_root), "pan-post-window");
  gtk_window_set_title (GTK_WINDOW(_root), _("Post Article"));
  g_object_set_data_full (G_OBJECT(_root), "post-ui", this, delete_post_ui);

  // populate the window
  GtkWidget * vbox = gtk_vbox_new (false, PAD_SMALL);
  GtkWidget * menu_vbox = gtk_vbox_new (false, PAD_SMALL);
  gtk_box_pack_start (GTK_BOX(vbox), menu_vbox, false, false, 0);
  add_actions (menu_vbox);
  gtk_window_add_accel_group (GTK_WINDOW(_root), gtk_ui_manager_get_accel_group (_uim));
  gtk_widget_show_all (vbox);
  gtk_container_add (GTK_CONTAINER(_root), vbox);

  // add the headers table
  const GtkAttachOptions fill ((GtkAttachOptions)(GTK_FILL|GTK_EXPAND|GTK_SHRINK));
  const GtkAttachOptions nofill ((GtkAttachOptions)0);
  char buf[512];
  int row = 0;
  GtkWidget *l, *t, *w;
  t = HIG :: workarea_create ();
  w = gtk_alignment_new (0.0f, 0.0f, 0.0f, 0.0f);
  gtk_widget_set_size_request (w, 12u, 0u);
  gtk_table_attach (GTK_TABLE(t), w, 2, 3, row, row+4, nofill, nofill, 0, 0);
  g_snprintf (buf, sizeof(buf), "<b>%s:</b>", _("F_rom"));
  l = HIG :: workarea_add_label (t, row, buf);
  w = _from_combo = gtk_combo_box_new_text ();
  gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
  g_signal_connect (w, "changed", G_CALLBACK(on_from_combo_changed), this);
  gtk_table_attach (GTK_TABLE(t), w, 3, 4, row, row+1, fill, nofill, 0, 0);
  ++row;
  g_snprintf (buf, sizeof(buf), "<b>%s:</b>", _("_Subject"));
  l = HIG :: workarea_add_label (t, row, buf);
  w = _subject_entry = gtk_entry_new ();
  gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
  gtk_table_attach (GTK_TABLE(t), w, 3, 4, row, row+1, fill, nofill, 0, 0);
  ++row;
  g_snprintf (buf, sizeof(buf), "<b>%s:</b>", _("_Newsgroups"));
  l = HIG :: workarea_add_label (t, row, buf);
  w = _groups_entry = gtk_entry_new ();
  gtk_label_set_mnemonic_widget (GTK_LABEL(l), w);
  gtk_table_attach (GTK_TABLE(t), w, 3, 4, row, row+1, fill, nofill, 0, 0);
  ++row;
  GtkWidget * frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
  GtkWidget * v = gtk_text_view_new ();
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW(v), false);
  gtk_tooltips_set_tip (tips, v, _("One header per line, in the form HeaderName: Value"), 0);
  _headers_buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(v));
  gtk_container_add (GTK_CONTAINER(frame), v);
  g_snprintf (buf, sizeof(buf), "<b>%s</b>", _("_More Headers"));
  w = GTK_WIDGET (g_object_new (GTK_TYPE_EXPANDER,
    "use-markup", TRUE,
    "use-underline", TRUE,
    "label", buf, NULL));
  gtk_container_add (GTK_CONTAINER(w), frame);
  gtk_table_attach (GTK_TABLE(t), w, 1, 4, row, row+1, fill, nofill, 0, 0);
  ++row;
  gtk_box_pack_start (GTK_BOX(vbox), t, false, false, 0);

  // add the body text widget
  w = create_body_widget (_body_buf, _body_view, prefs);
  g_signal_connect (_body_buf, "insert-text", G_CALLBACK(text_was_inserted_cb), &_tm);
  gtk_box_pack_start (GTK_BOX(vbox), w, true, true, 0);

  // wire ourselves up with the parent
  if (parent != 0) {
    gtk_window_set_transient_for (GTK_WINDOW(_root), parent);
    gtk_window_set_position (GTK_WINDOW(_root), GTK_WIN_POS_CENTER_ON_PARENT);
  }

  // ensure that we've got at least one profile to choose from...
  gtk_widget_show_all (vbox);
  if (!_profiles.has_profiles()) {
    gtk_widget_show_all (_root);
    Profile profile;
    profile.username = g_get_real_name ();
    profile.attribution = _("On %d, %n wrote:");
    ProfileDialog d (g_get_real_name(), profile, GTK_WINDOW(_root));
    ProfileDialog :: run_until_valid (d);
    std::string name;
    d.get_profile (name, profile);
    _profiles.add_profile (name, profile);
    gtk_widget_destroy (d.root());
  }

  set_message (message);

  // set focus to the first non-populated widget
  GtkWidget * grab (0);
  if (!grab) {
    const char * cpch (gtk_entry_get_text (GTK_ENTRY(_subject_entry)));
    if (!cpch || !*cpch)
      grab = _subject_entry;
  }
  if (!grab) {
    const char * cpch (gtk_entry_get_text (GTK_ENTRY(_groups_entry)));
    if (!cpch || !*cpch)
      grab = _groups_entry;
  }
  if (!grab)
    grab = _body_view;
  gtk_widget_grab_focus (grab);
}
