/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*********************
**********************  Includes
*********************/

#include <config.h>
#include <gtk/gtk.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/usenet-utils/gnksa.h>
#include "hig.h"
#include "pad.h"
#include "profiles-dialog.h"

using namespace pan;

/***
****
***/

namespace
{
  void set_entry (GtkWidget * w, const std::string& str)
  {
    const char * s = str.empty() ? "" : str.c_str();
    gtk_entry_set_text (GTK_ENTRY(w), s);
  }

  void from_entry (GtkWidget * w, std::string& setme)
  {
    StringView v (gtk_entry_get_text (GTK_ENTRY(w)));
    v.trim ();
    setme.assign (v.str, v.len);
  }

  void on_sig_file_toggled (GtkToggleButton * tb, gpointer sensitize)
  {
    gtk_widget_set_sensitive (GTK_WIDGET(sensitize), gtk_toggle_button_get_active(tb));
  }
}

ProfileDialog :: ProfileDialog (const StringView& profile_name,
                                const Profile& profile,
                                GtkWindow* parent)
{
  GtkTooltips * ttips = gtk_tooltips_new ();

  _root = gtk_dialog_new_with_buttons (_("Posting Profile"), parent,
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
                                       NULL);
  gtk_window_set_role (GTK_WINDOW(_root), "pan-edit-profile-dialog");

  int row (0);
  GtkWidget *t = HIG :: workarea_create ();
  HIG :: workarea_add_section_title (t, &row, _("Profile Information"));
    HIG :: workarea_add_section_spacer (t, row, 1);
    GtkWidget * w = _name_entry = gtk_entry_new ();
    set_entry (w, profile_name);
    HIG :: workarea_add_row (t, &row, _("_Profile Name:"), w);

  HIG :: workarea_add_section_divider (t, &row);
  HIG :: workarea_add_section_title (t, &row, _("Required Information"));
    HIG :: workarea_add_section_spacer (t, row, 2);
    w = _username_entry = gtk_entry_new ();
    set_entry (w, profile.username);
    HIG :: workarea_add_row (t, &row, _("_Full Name:"), w);
    w = _address_entry = gtk_entry_new ();
    set_entry (w, profile.address);
    HIG :: workarea_add_row (t, &row, _("_Email Address:"), w);

  HIG :: workarea_add_section_divider (t, &row);
  HIG :: workarea_add_section_title (t, &row, _("Optional Information"));
    HIG :: workarea_add_section_spacer (t, row, 5);
    w = _reply_to_entry = gtk_entry_new ();
    set_entry (w, profile.get_header("Reply-To"));
    gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("Used when email replies to your posts should go somewhere other than the email address above.  Because Reply-To: is relatively difficult for spammers to harvest, it's polite give your email address here if you gave an invalid email address above."), "");
    HIG :: workarea_add_row (t, &row, _("_Reply-To Header:"), w, NULL);
    w = _organization_entry = gtk_entry_new ();
    set_entry (w, profile.get_header("Organization"));
    HIG :: workarea_add_row (t, &row, _("Or_ganization Header:"), w, NULL);
    //w = _msgid_fqdn_entry = gtk_entry_new ();
    //gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("When posting to Usenet, your article's Message-ID contains a domain name.  You can set a custom domain name here, or leave it blank to let Pan use the domain name from your email address."), "");
    //HIG :: workarea_add_row (t, &row, _("Message-ID _Domain Name:"), w, NULL);
    w = _attribution_entry = gtk_entry_new ();
    set_entry (w, profile.attribution);
    gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w, _("%i for Message-ID\n%a for Author and Address\n%n for Author name\n%d for Date"), "");
    HIG :: workarea_add_row (t, &row, _("_Attribution:"), w, NULL);
    GtkWidget * toggle = _signature_file_check = gtk_check_button_new_with_mnemonic (_("Attach this _Signature:"));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(toggle), profile.use_sigfile);
    gtk_table_attach (GTK_TABLE(t), toggle, 1, 2, row, row+1, GTK_FILL, (GtkAttachOptions)0, 0, 0);
#if GTK_CHECK_VERSION(2,6,0)
    w = _signature_file_control = gtk_file_chooser_button_new (_("Select Signature File"), GTK_FILE_CHOOSER_ACTION_OPEN);
    if (!profile.signature_file.empty())
      gtk_file_chooser_set_filename (GTK_FILE_CHOOSER(w), profile.signature_file.c_str());
#else
    w = _signature_file_control = gtk_entry_new ();
    set_entry (w, profile.signature_file);
#endif
    g_signal_connect (_signature_file_check, "toggled", G_CALLBACK(on_sig_file_toggled), _signature_file_control);
    HIG :: workarea_add_control (t, row, w);
    ++row;

  on_sig_file_toggled (GTK_TOGGLE_BUTTON(_signature_file_check), _signature_file_control);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(_root)->vbox), t, true, true, 0);
  gtk_widget_show_all (t);

  if (parent != 0) {
    gtk_window_set_transient_for (GTK_WINDOW(_root), parent);
    gtk_window_set_position (GTK_WINDOW(_root), GTK_WIN_POS_CENTER_ON_PARENT);
  }
}

ProfileDialog :: ~ProfileDialog ()
{
}

namespace
{
  void message_dialog_set_text (GtkMessageDialog * dialog,
                                const char * primary,
                                const char * secondary)
  {
#if GTK_CHECK_VERSION(2,6,0)
    gtk_message_dialog_set_markup (dialog, primary);
    gtk_message_dialog_format_secondary_text (dialog, "%s", secondary);
#else
    char * pch = g_strdup_printf ("<b>%s</b>\n\n%s", primary, secondary);
    gtk_message_dialog_set_markup (dialog, pch);
    g_free (pch);
#endif
  }
}

/*static*/ void
ProfileDialog :: run_until_valid (ProfileDialog& pd)
{
  for (;;)
  {
    gtk_dialog_run (GTK_DIALOG(pd.root()));

    std::string name;
    Profile profile;
    pd.get_profile (name, profile);

    if (GNKSA :: check_from (profile.address, true))
    {
      GtkWidget * d = gtk_message_dialog_new (
        GTK_WINDOW(pd.root()),
        GTK_DIALOG_DESTROY_WITH_PARENT,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_CLOSE,
        NULL);
      message_dialog_set_text (GTK_MESSAGE_DIALOG(d),
                               _("Invalid email address."),
                               _("Please use an address of the form joe@somewhere.org"));
      gtk_dialog_run (GTK_DIALOG(d));
      gtk_widget_destroy (d);
      gtk_widget_grab_focus (pd._address_entry);
    }
    else break;
  }
}

void
ProfileDialog :: get_profile (std::string& profile_name, Profile& profile)
{
  from_entry (_name_entry, profile_name);
  from_entry (_username_entry, profile.username);
  from_entry (_address_entry, profile.address);
  from_entry (_attribution_entry, profile.attribution);

  profile.use_sigfile = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(_signature_file_check));
#if GTK_CHECK_VERSION(2,6,0)
  char * pch = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(_signature_file_control));
  profile.signature_file.clear ();
  if (pch && *pch)
    profile.signature_file = pch;
  g_free (pch);
#else
  from_entry (_signature_file_control, profile.signature_file);
#endif

  std::string s;
  from_entry (_reply_to_entry, s);
  if (s.empty())
    profile.headers.erase ("Reply-To");
  else
    profile.headers["Reply-To"] = s;

  from_entry (_organization_entry, s);
  if (s.empty())
    profile.headers.erase ("Organization");
  else
    profile.headers["Organization"] = s;
}

/******
*******
*******
******/

namespace
{
  enum { COL_NAME, N_COLS };

  GtkListStore* new_profile_store (const Profiles& profiles)
  {
    GtkListStore * store = gtk_list_store_new (N_COLS, G_TYPE_STRING);
    typedef std::set<std::string> names_t;
    const names_t profile_names (profiles.get_profile_names ());
    foreach_const (names_t, profile_names, it) {
      GtkTreeIter iter;
      gtk_list_store_append (store, &iter);
      gtk_list_store_set (store, &iter, COL_NAME, it->c_str(), -1);
    }
    return store;
  }

  std::string get_selected_profile (GtkTreeView * view)
  {
    std::string name;
    GtkTreeSelection * selection (gtk_tree_view_get_selection (view));
    GtkTreeModel * model;
    GtkTreeIter iter;
    if (gtk_tree_selection_get_selected (selection, &model, &iter)) {
      char * pch (0);
      gtk_tree_model_get (model, &iter, COL_NAME, &pch, -1);
      if (pch) {
        name = pch;
        g_free (pch);
      }
    }
    return name;
  }

  static void on_add_button (GtkWidget*, gpointer d) { static_cast<ProfilesDialog*>(d)->create_new_profile(); }
  static void on_edit_button (GtkWidget*, gpointer d) { static_cast<ProfilesDialog*>(d)->edit_profile(); }
  static void on_delete_button (GtkWidget*, gpointer d) { static_cast<ProfilesDialog*>(d)->delete_profile(); }
}

void
ProfilesDialog :: rebuild_store ()
{
  _store = new_profile_store (_profiles);
  gtk_tree_view_set_model (GTK_TREE_VIEW(_view), GTK_TREE_MODEL(_store));
  g_object_unref (_store);
}

void
ProfilesDialog :: delete_profile ()
{
  const std::string name (get_selected_profile (GTK_TREE_VIEW (_view)));
  _profiles.delete_profile (name);
  rebuild_store ();
  refresh_buttons ();
}

void
ProfilesDialog :: edit_profile ()
{
  const std::string old_name (get_selected_profile (GTK_TREE_VIEW (_view)));
  if (!old_name.empty())
  {
    Profile profile;
    if (_profiles.get_profile (old_name, profile))
    {
      ProfileDialog d (old_name, profile, GTK_WINDOW(_root));
      ProfileDialog :: run_until_valid (d);

      std::string new_name;
      d.get_profile (new_name, profile);
      _profiles.delete_profile (old_name);
      _profiles.add_profile (new_name, profile);
      rebuild_store ();
      refresh_buttons ();

      gtk_widget_destroy (d.root());
    }
  }
}

void
ProfilesDialog :: create_new_profile ()
{
  Profile profile;
  profile.username = g_get_real_name ();
  profile.attribution = _("On %d, %n wrote:");
  ProfileDialog d (_("New Profile"), profile, GTK_WINDOW(_root));
  ProfileDialog :: run_until_valid (d);
  std::string name;
  d.get_profile (name, profile);
  _profiles.add_profile (name, profile);
  gtk_widget_destroy (d.root());
  rebuild_store ();
  refresh_buttons ();
}

void
ProfilesDialog :: refresh_buttons ()
{
  const std::string name (get_selected_profile (GTK_TREE_VIEW (_view)));
  bool have_sel (!name.empty());
  gtk_widget_set_sensitive (_edit_button, have_sel);
  gtk_widget_set_sensitive (_remove_button, have_sel);
}

namespace
{
  static void profiles_tree_view_selection_changed_cb (GtkTreeSelection * selection, gpointer user_data)
  {
    ProfilesDialog *d = (ProfilesDialog *)user_data;
    d->refresh_buttons ();
  }
}

ProfilesDialog :: ~ProfilesDialog ()
{
}

ProfilesDialog :: ProfilesDialog (Profiles &profiles, GtkWindow *parent):
  _profiles (profiles)
{
  _root = gtk_dialog_new_with_buttons (_("Posting Profiles"), parent,
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
                                       NULL);
  gtk_window_set_role (GTK_WINDOW(_root), "pan-profiles-dialog");
  //g_signal_connect (GTK_OBJECT(_root), "response", G_CALLBACK(response_cb), this);

  // workarea
  GtkWidget * hbox = gtk_hbox_new (false, PAD);
  gtk_container_set_border_width (GTK_CONTAINER(hbox), PAD_BIG);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(_root)->vbox), hbox, true, true, 0);

  // create the list
  GtkWidget * w = _view = gtk_tree_view_new ();
  rebuild_store ();
  GtkCellRenderer * renderer = gtk_cell_renderer_text_new ();
  GtkTreeViewColumn * column = gtk_tree_view_column_new_with_attributes (_("Profiles"), renderer, "text", COL_NAME, NULL);
  gtk_tree_view_column_set_sort_column_id (column, COL_NAME);
  gtk_tree_view_append_column (GTK_TREE_VIEW(w), column);
  GtkTreeSelection * selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (w));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
  g_signal_connect (G_OBJECT (selection), "changed",
                    G_CALLBACK (profiles_tree_view_selection_changed_cb), this);
  w = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(w), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(w), GTK_SHADOW_IN);
  gtk_container_add (GTK_CONTAINER(w), _view);
  gtk_box_pack_start (GTK_BOX (hbox), w, true, true, 0);
  gtk_widget_set_size_request (w, 300, 300);

  // button box
  GtkTooltips * tips = gtk_tooltips_new ();
  GtkWidget * bbox = gtk_vbox_new (FALSE, PAD_SMALL);
  gtk_box_pack_start (GTK_BOX (hbox), bbox, false, false, 0);

  // add button
  w = gtk_button_new_from_stock (GTK_STOCK_ADD);
  gtk_box_pack_start (GTK_BOX (bbox), w, false, false, 0);
  gtk_tooltips_set_tip (tips, w, _("Add a New Profile"), NULL);
  gtk_signal_connect (GTK_OBJECT(w), "clicked", GTK_SIGNAL_FUNC(on_add_button), this);

  // edit button
#if GTK_CHECK_VERSION(2,6,0)
  w = gtk_button_new_from_stock (GTK_STOCK_EDIT);
#else
  w = gtk_button_new_from_stock (GTK_STOCK_OPEN);
#endif
  _edit_button = w;
  gtk_box_pack_start (GTK_BOX (bbox), w, false, false, 0);
  gtk_tooltips_set_tip (tips, w, _("Edit the selected profile"), NULL);
  gtk_signal_connect (GTK_OBJECT(w), "clicked", GTK_SIGNAL_FUNC(on_edit_button), this);

  // remove button
  _remove_button = w = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
  gtk_box_pack_start (GTK_BOX (bbox), w, false, false, 0);
  gtk_tooltips_set_tip (tips, w, _("Remove the selected profile"), NULL);
  gtk_signal_connect (GTK_OBJECT(w), "clicked", GTK_SIGNAL_FUNC(on_delete_button), this);

  // set sensitive buttons
  refresh_buttons ();

  gtk_widget_show_all (hbox);

  if (parent != 0) {
    gtk_window_set_transient_for (GTK_WINDOW(_root), parent);
    gtk_window_set_position (GTK_WINDOW(_root), GTK_WIN_POS_CENTER_ON_PARENT);
  }
}
