//* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/usenet-utils/utf8-utils.h>
#include <pan/tasks/queue.h>
#include <pan/icons/pan-pixbufs.h>
#include "pad.h"
#include "task-pane.h"

enum
{
  COL_TASK_POINTER,
  COL_TASK_STATE,
  NUM_COLS
};


/**
***  Internal Utility
**/

typedef Queue::task_list_t task_list_t;

void
TaskPane :: get_selected_tasks_foreach (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer list_g)
{
  Task * task (0);
  gtk_tree_model_get (model, iter, COL_TASK_POINTER, &task, -1);
  static_cast<task_list_t*>(list_g)->push_back (task);
}

task_list_t
TaskPane :: get_selected_tasks () const
{
  task_list_t tasks;
  GtkTreeView * view (GTK_TREE_VIEW (_view));
  GtkTreeSelection * sel (gtk_tree_view_get_selection (view));
  gtk_tree_selection_selected_foreach (sel, get_selected_tasks_foreach, &tasks);
  return tasks;
}

/**
***  User Interactions
**/

namespace
{
  int
  find_task_index (GtkListStore * list, Task * task)
  {
    GtkTreeIter iter;
    int index (0);

    if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list), &iter)) do
    {
      Task * current_task;
      gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COL_TASK_POINTER, &current_task, -1);
      if (current_task == task)
        return index;
      ++index;
    }
    while (gtk_tree_model_iter_next (GTK_TREE_MODEL(list), &iter));

    // not found
    return -1;
  }
}

void TaskPane :: up_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.move_up (pane->get_selected_tasks());
}
void TaskPane :: down_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.move_down (pane->get_selected_tasks());
}
void TaskPane :: top_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.move_top (pane->get_selected_tasks());
}
void TaskPane :: bottom_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.move_bottom (pane->get_selected_tasks());
}
void TaskPane :: stop_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.stop_tasks (pane->get_selected_tasks());
}
void TaskPane :: delete_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.remove_tasks (pane->get_selected_tasks());
}

void TaskPane :: restart_clicked_cb (GtkToolButton* b, TaskPane* pane)
{
  pane->_queue.restart_tasks (pane->get_selected_tasks());
}

/**
***  Display
**/

namespace
{
  typedef Queue::tasks_t task_states_t;

  void
  add_tasks_to_list (GtkListStore        * store,
                     Queue               & queue,
                     task_list_t         & tasks,
                     int                   index)
  {
    task_states_t task_states;
    queue.get_tasks (task_states);

    if (index == -1)
      index = gtk_tree_model_iter_n_children (GTK_TREE_MODEL(store), NULL);

    GtkTreeIter iter;
    foreach (task_list_t, tasks, it) {
      Task * task (*it);
      gtk_list_store_insert (store, &iter, index++);
      gtk_list_store_set (store, &iter, COL_TASK_POINTER, task,
                                        COL_TASK_STATE, (int)task_states.get_state(task),
                                        -1);
    }
  }
}

/***
****  Queue Listener
***/

void
TaskPane :: on_queue_tasks_added (Queue& queue, std::vector<Task*>& tasks, int index)
{
  add_tasks_to_list (_store, _queue, tasks, index);
}

void
TaskPane :: on_queue_task_removed (Queue& queue, Task& task, int index)
{
  const int list_index (find_task_index (_store, &task));
  assert (list_index == index);
  GtkTreeIter iter;
  gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(_store), &iter, NULL, index);
  gtk_list_store_remove (_store, &iter);
}

void
TaskPane :: on_queue_task_moved (Queue& queue, Task& task, int new_index, int old_index)
{
  std::cerr << LINE_ID << " FIXME" << std::endl;
  const int old_list_index (find_task_index (_store, &task));
  assert (old_list_index == old_index);
  GtkTreeIter task_iter, before_me;
  gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(_store), &task_iter, NULL, old_index);
  gtk_tree_model_iter_nth_child (GTK_TREE_MODEL(_store), &before_me, NULL, new_index);
  gtk_list_store_move_before (_store, &task_iter, &before_me);
}

void TaskPane :: on_queue_task_active_changed (Queue&, Task&, bool active) { }
void TaskPane :: on_queue_connection_count_changed (Queue&, int count) { }
void TaskPane :: on_queue_size_changed (Queue&, int active, int total) { }
void TaskPane :: on_queue_online_changed (Queue&, bool online) { }


/***
****
***/

void
TaskPane :: update_titlebar (const task_states_t& tasks)
{
  int queued_count (0);
  int stopped_count (0);
  int running_count (0);

  foreach_const (task_states_t::states_t, tasks.states, it)
  {
    const Queue::TaskState state (it->second);
    if (state == Queue::RUNNING)
      ++running_count;
    else if (state == Queue::STOPPED)
      ++stopped_count;
    else if (state == Queue::QUEUED)
      ++queued_count;
  }

  char title[1024];
  if (stopped_count)
    g_snprintf (title, sizeof(title), _("Pan: Tasks (%d Queued, %d Running, %d Stopped)"), queued_count, running_count, stopped_count);
  else if (running_count || queued_count)
    g_snprintf (title, sizeof(title), _("Pan: Tasks (%d Queued, %d Running)"), queued_count, running_count);
  else 
    g_snprintf (title, sizeof(title), _("Pan: Tasks"));

  gtk_window_set_title (GTK_WINDOW(_root), title);
}

gboolean
TaskPane :: periodic_refresh_foreach (GtkTreeModel    * model,
                                      GtkTreePath     * path,
                                      GtkTreeIter     * iter,
                                      gpointer          states_gpointer)
{
  Task * task;
  int old_state;
  gtk_tree_model_get (model, iter, COL_TASK_POINTER, &task,
                                   COL_TASK_STATE, &old_state, -1);

  const task_states_t * states (static_cast<const task_states_t*>(states_gpointer));
  const int new_state (states->get_state (task));

  if (new_state != old_state)
    gtk_list_store_set (GTK_LIST_STORE(model), iter, COL_TASK_POINTER, task,
                                                     COL_TASK_STATE, new_state,
                                                     -1);

  return false; // keep iterating
}

gboolean
TaskPane :: periodic_refresh (gpointer pane_gpointer)
{
  TaskPane * pane (static_cast<TaskPane*>(pane_gpointer));
  task_states_t tasks;
  pane->_queue.get_tasks (tasks);
  gtk_tree_model_foreach (GTK_TREE_MODEL(pane->_store), periodic_refresh_foreach, &tasks);
  pane->update_titlebar (tasks);
  gtk_widget_queue_draw (pane->_view);

  return true;
}
#if 0
static gint
window_delete_event_cb (GtkWidget * w, GdkEvent * e, gpointer data)
{
	PaneUI * ui = (PaneUI*) data;
	gui_save_window_size (ui->window, "task_manager_2");
	gui_save_column_widths_tree_view (ui->task_manager_tree_view, "task_manager_2");
	return FALSE;
} 
#endif

namespace
{
  void
  render_state (GtkTreeViewColumn *col,
                GtkCellRenderer *rend,
                GtkTreeModel *model,
                GtkTreeIter *iter,
                Queue * queue)
  {
    Task * task (0);
    int state;
    gtk_tree_model_get (model, iter, COL_TASK_POINTER, &task,
                                     COL_TASK_STATE,  &state,
                                     -1);

    // build the state string
    const char * state_str (0);
    switch (state) {
      case Queue::RUNNING:  state_str = _("Running"); break;
      case Queue::QUEUED:   state_str = _("Queued"); break;
      case Queue::STOPPED:  state_str = _("Stopped"); break;
      case Queue::REMOVING: state_str = _("Removing"); break;
      default: state_str = _("Unknown"); break;
    }

    // get the time remaining
    const unsigned long bytes_remaining (task->get_bytes_remaining ());
    double speed;
    int connections;
    queue->get_task_speed_KiBps (task, speed, connections);
    const unsigned long seconds_remaining (speed>0.001 ? (unsigned long)(bytes_remaining/(speed*1024)) : 0);
    int h(0), m(0), s(0);
    if (seconds_remaining > 0) {
      h = seconds_remaining / 3600;
      m = (seconds_remaining % 3600) / 60;
      s = seconds_remaining % 60;
    }

    // get the percent done
    const int percent (task->get_progress_of_100());

    // get the description
    const std::string description (task->describe ());

    char line2[1024];
    if (state == Queue::RUNNING)
      g_snprintf (line2, sizeof(line2), _("%d%% Done; %d:%02d:%02d Remaining (%d @ %lu KiB/s)"), percent, h, m, s, connections, (unsigned long)speed);
    else if (!percent)
      *line2 = '\0';
    else
      g_snprintf (line2, sizeof(line2), _("%d%% Done; %s"), percent, state_str);
    char * str = g_markup_printf_escaped ("%s\n<small>%s</small>", description.c_str(), line2);
    char * str_utf8 = pan::clean_utf8 (str, -1);
    g_object_set(rend, "markup", str_utf8, "xpad", 10, "ypad", 5, NULL);
    g_free (str_utf8);
    g_free (str);
  }
}

/**
***
**/

namespace
{
  void delete_task_pane (gpointer pane_gpointer)
  {
    delete static_cast<TaskPane*>(pane_gpointer);
  }
}

TaskPane :: ~TaskPane ()
{
  _queue.remove_listener (this);
  g_source_remove (_update_timeout_tag);
}

TaskPane :: TaskPane (Queue& queue, GtkWidget* parent): _queue(queue)
{
  _root = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_role (GTK_WINDOW(_root), "pan-tasks-window");
  GdkPixbuf * pixbuf = gdk_pixbuf_new_from_inline (-1, icon_pan, FALSE, NULL);
  gtk_window_set_icon (GTK_WINDOW(_root), pixbuf);
  g_object_unref (pixbuf);

  GtkTooltips * ttips = gtk_tooltips_new ();

  GtkWidget * vbox = gtk_vbox_new (false, PAD_SMALL);

  GtkWidget * toolbar = gtk_toolbar_new ();

    GtkToolItem * item = gtk_tool_button_new_from_stock (GTK_STOCK_GO_UP);
    g_signal_connect (item, "clicked", G_CALLBACK(up_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_GOTO_TOP);
    g_signal_connect (item, "clicked", G_CALLBACK(top_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_separator_tool_item_new ();
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_GO_DOWN);
    g_signal_connect (item, "clicked", G_CALLBACK(down_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_GOTO_BOTTOM);
    g_signal_connect (item, "clicked", G_CALLBACK(bottom_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_separator_tool_item_new ();
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_REDO);
    gtk_tool_item_set_tooltip (item, ttips, _("Restart Tasks"), 0);
    g_signal_connect (item, "clicked", G_CALLBACK(restart_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_STOP);
    gtk_tool_item_set_tooltip (item, ttips, _("Stop Tasks"), 0);
    g_signal_connect (item, "clicked", G_CALLBACK(stop_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_DELETE);
    gtk_tool_item_set_tooltip (item, ttips, _("Delete Tasks"), 0);
    g_signal_connect (item, "clicked", G_CALLBACK(delete_clicked_cb), this);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_separator_tool_item_new ();
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE);
    g_signal_connect_swapped (item, "clicked", G_CALLBACK(gtk_widget_destroy), _root);
    gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

  gtk_box_pack_start (GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);

  _store = gtk_list_store_new (NUM_COLS, G_TYPE_POINTER, G_TYPE_INT);
  _view = gtk_tree_view_new_with_model (GTK_TREE_MODEL(_store));
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW(_view), false);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW(_view), true);
  GtkCellRenderer * renderer = gtk_cell_renderer_text_new ();
  GtkTreeViewColumn * col = gtk_tree_view_column_new_with_attributes (_("Status"), renderer, NULL);
  gtk_tree_view_column_set_cell_data_func (col, renderer, (GtkTreeCellDataFunc)render_state, &_queue, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_view), col);
  GtkTreeSelection * selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (_view));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);

  GtkWidget * w = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(w), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(w), GTK_SHADOW_IN);
  gtk_container_add (GTK_CONTAINER(w), _view);
  gtk_box_pack_start (GTK_BOX(vbox), w, true, true, 0);

  // populate the list
  task_states_t tasks;
  _queue.get_tasks (tasks);
  add_tasks_to_list (_store, _queue, tasks.tasks, 0);
  _queue.add_listener (this);

  _update_timeout_tag = g_timeout_add (750, periodic_refresh, this);

  gtk_window_set_default_size (GTK_WINDOW(_root), 525, 600);
  g_object_set_data_full (G_OBJECT(_root), "pane", this, delete_task_pane);
  gtk_container_add (GTK_CONTAINER(_root), vbox);
  gtk_widget_show_all (vbox);

  if (parent != 0) {
    gtk_window_set_transient_for (GTK_WINDOW(_root), GTK_WINDOW(gtk_widget_get_toplevel(parent)));
    gtk_window_set_position (GTK_WINDOW(_root), GTK_WIN_POS_CENTER_ON_PARENT);
  }
}
