/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/messages.h>
#include "nntp-pool.h"

using namespace pan;

NNTP_Pool :: NNTP_Pool (const Quark        & server,
                        Socket::Creator    * creator,
                        const std::string  & username,
                        const std::string  & password):
  _server (server),
  _socket_creator (creator),
  _is_online (true),
  _connection_pending (false),
  _port (119),
  _username (username),
  _password (password),
  _max_connections (4)
{
}

NNTP_Pool :: ~NNTP_Pool ()
{
  foreach (pool_items_t, _pool_items, it) {
    delete it->nntp->_socket;
    delete it->nntp;
  }
}

/***
****
***/

void
NNTP_Pool :: set_address (const StringView& host, int port)
{
   _host = host.to_string ();
   _port = port;
}

void
NNTP_Pool :: set_auth (const StringView& username, const StringView& password)
{
std::cerr <<__FILE__ <<':'<<__LINE__ << " nntp_pool set auth username [" << username << "] password [" << password << "]\n";
   _username = username.to_string ();
   _password = password.to_string ();
}

void
NNTP_Pool :: set_max_connections (int max_connections)
{
   _max_connections = max_connections;
}

void
NNTP_Pool :: set_online (bool b)
{
  if (b != _is_online)
  {
    _is_online = b;

    if (!_is_online)
      foreach (pool_items_t, _pool_items, it)
        if (!it->is_checked_in)
          it->nntp->_socket->set_abort_flag (true);
  }
}

bool
NNTP_Pool :: is_online () const
{
  return _is_online;
}

NNTP*
NNTP_Pool :: check_out ()
{
  NNTP * nntp (0);

  foreach (pool_items_t, _pool_items, it) {
    if (it->is_checked_in) {
      nntp = it->nntp;
      it->is_checked_in = false;
      debug ("nntp " << nntp << " is now checked out");
      break;
    }
  }

  return nntp;
}

void
NNTP_Pool :: check_in (NNTP * nntp, bool is_ok)
{
  debug ("nntp " << nntp << " is being checked in, is_ok is " << is_ok);

  // find this nntp in _pool_items
  pool_items_t::iterator it;
  for (it=_pool_items.begin(); it!=_pool_items.end(); ++it)
    if (it->nntp == nntp)
      break;

  // process the nntp if we have a match
  if (it != _pool_items.end()) {
    if (is_ok) {
      it->is_checked_in = true;
      it->last_active_time = time (NULL);
      fire_nntp_available ();
    } else {
      delete it->nntp->_socket;
      delete it->nntp;
      _pool_items.erase (it);
    }
  }
}

/***
****
***/

void
NNTP_Pool :: add_listener (Listener * l)
{
   _listeners.insert (l);
}

void
NNTP_Pool :: remove_listener (Listener * l)
{
   _listeners.erase (l);
}

void
NNTP_Pool :: on_nntp_done (NNTP* nntp, Health health)
{
   debug ("NNTP_Pool: on_nntp_done()");

   const Quark& server (nntp->_server);

   if (health != OK)
   {
      delete nntp->_socket;
      delete nntp;
      nntp = 0;
   }

   if (health == FAIL) // news server isn't accepting our connection!
      set_online (false);

   _connection_pending = false;

   // if success...
   if (nntp != 0)
   {
      debug ("success with handshake to " << server.to_string() << ", nntp " << (void*)nntp);

      PoolItem i;
      i.nntp = nntp;
      i.is_checked_in = true;
      i.last_active_time = time (0);
      _pool_items.push_back (i);

      fire_nntp_available ();
   }
}

void
NNTP_Pool :: fire_nntp_available ()
{
  for (listeners_t::iterator it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_nntp_available (_server);
}

void
NNTP_Pool :: get_counts (int& setme_active, int& setme_idle, int& setme_pending, int& setme_max) const
{
  int idle (0);
  int active (0);

  foreach_const (pool_items_t, _pool_items, it)
    if (it->is_checked_in)
      ++idle;
    else
      ++active;

  setme_active = active;
  setme_idle = idle;
  setme_max = _max_connections;
  setme_pending  = _connection_pending ? 1 : 0;
}


void
NNTP_Pool :: request_nntp ()
{
   int active, idle, pending, max;
   get_counts (active, idle, pending, max);

   debug ("server " << _server << ", is_online(): " << is_online() << ' '
          << "active: " << active << ' '
          << "idle: " << idle << ' '
          << "pending: " << pending << ' '
          << "max: " << max << ' ');

   if (is_online()
       && !idle
       && !pending
       && active<max)
   {
      debug ("trying to create a socket");

      Socket * sock = _socket_creator->create_socket (_host, _port);
      if (sock != 0)
      {
         _connection_pending = true;
         NNTP * nntp = new NNTP (_server, _username, _password, sock);
         nntp->handshake (this);
      }
   }
}
