/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <ctype.h>
#include <errno.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>

#include <glib.h>

#include <gmime/gmime.h>
#include <gmime/gmime-stream-file.h>

#include <pan/general/debug.h>
#include <pan/general/i18n.h>
#include <pan/general/log.h>
#include <pan/general/file-util.h>

#include "decode.h"
#include "mime-utils.h"

#define is_nonempty_string(a) ((a) && (*a))

using namespace pan;

/**
*** Private Routines
**/

static gchar*
create_filename (const gchar       * path,
                 const gchar       * default_filename,
                 const GMimePart   * part)
{
	GString * filename;
	char * retval;

	g_return_val_if_fail (is_nonempty_string(path), NULL);
	g_return_val_if_fail (part!=NULL, NULL);

       	filename = g_string_new (NULL);

	/* first try the filename specified by the user */
	if (!filename->len && is_nonempty_string(default_filename))
		g_string_assign (filename, default_filename);

	/* otherwise try the filename specified by the article */
	if (!filename->len) {
		const gchar * pch = g_mime_part_get_filename (part);
		if (is_nonempty_string(pch)) {
			/* if some bozo included the windows pathname in the filename, strip that out */
			if (isalpha((guchar)pch[0]) && pch[1]==':' && pch[2]=='\\')
				pch = strrchr (pch, '\\') + 1;
			g_string_assign (filename, pch);
		}
	}

	/* otherwise punt */
	if (!filename->len)
		g_string_assign (filename, _("UNKNOWN"));

	retval = file :: get_unique_fname (path, filename->str);
	g_string_free (filename, TRUE);

	return retval;
}


/*****
******
*****/

static void
get_array_of_decodable_parts_func (GMimeObject *part, gpointer data)
{
	GPtrArray * a = (GPtrArray*) data;
	const GMimeContentType * type;
	
	if (GMIME_IS_MULTIPART (part))
		return;

	type = g_mime_object_get_content_type (part);
	if ((is_nonempty_string (g_mime_part_get_filename (GMIME_PART (part)))) ||
	    (!g_mime_content_type_is_type (type, "text", "*")))
		g_ptr_array_add (a, part);
}


/**
*** Public Routines
**/

bool
Decode :: decode (const Quark                   & path,
                  const Quark                   & suggested_filename,
                  GMimeMessage                  * mm,
                  Progress                      & progress)
{
	GPtrArray * attachments = NULL;
	int i = 0;
	gboolean success = TRUE;
	char * filename = NULL;
	GMimeStream * stream;
	GMimeStream * stream_file = NULL;
	GMimeFilterYenc * yenc_filter;

	/* sanity clause */
	g_return_val_if_fail (mm!=0, false);

	/* get the message from the cache, and make an array of its attachments */
	attachments = g_ptr_array_new ();
	if (mm != NULL)
		g_mime_message_foreach_part (mm, get_array_of_decodable_parts_func, attachments);

	/* decode & save the parts */
	success = attachments->len != 0;
	stream = NULL;
	yenc_filter = NULL;
	for (i=0; success && i!=(int)attachments->len; ++i)
	{
		const GMimePart * part = (const GMimePart*) g_ptr_array_index (attachments, i);
		GMimeDataWrapper * content;
		mime::YencInfo * yenc = (mime::YencInfo*) g_object_get_data (G_OBJECT(part), "yenc");
		int part_number = -1;

		if (yenc != NULL)
			part_number = yenc->part;

		if (part_number < 2)
		{
			g_free (filename);
			filename = create_filename (path.c_str(), suggested_filename.c_str(), part);
			debug ("using filename [" << filename << "]");

			/* access the path */
			if (success) {
				char * dir = g_path_get_dirname (filename);
				bool dir_ok = file :: ensure_dir_exists (dir);
				if (!dir_ok) {
					Log::add_err_va (_("Decode can't access directory \"%s\""), dir);
					success = FALSE;
				}
				g_free (dir);
			}

			/* open the output file */
			if (success) {
				errno = 0;
				FILE * fp = fopen (filename, "wb+");
				debug ("writing to file [" << filename << "]");
				if (stream != NULL) {
					g_object_unref (G_OBJECT(stream));
					stream = 0;
				}
				if (fp != NULL) {
					stream = stream_file = g_mime_stream_file_new (fp);
					if (yenc != NULL) {
						GMimeStream * inner_stream = stream;
						GMimeFilter * filter;
						stream = g_mime_stream_filter_new_with_stream (inner_stream);
						filter = g_mime_filter_yenc_new (GMIME_FILTER_YENC_DIRECTION_DECODE);
						g_mime_stream_filter_add (GMIME_STREAM_FILTER(stream), filter);
						yenc_filter = (GMimeFilterYenc*) filter;
						g_object_unref (filter); // owned by `stream' now
						g_object_unref (inner_stream); // owned by `stream' now
					}
				}
				else {
				        Log::add_err_va (_("Unable to save \"%s\" %s"),
					                 filename,
                                                         file::pan_strerror(errno));
					success = FALSE;
				}
			}
		}

		/* make sure we have the necessary info to continue */          
		if (success) {
			if (stream == NULL || filename == NULL ||
			    (yenc != NULL && yenc_filter == NULL)) { 
				Log::add_err_va (_("Could not decode article - file may be corrupt/incomplete"));
				success = FALSE;
			}
		}
		
		/* make sure there's content to write */
		content = g_mime_part_get_content_object (part);
		if (!content)
			success = FALSE;

		/* write the content */
		if (success)
		{
			/* write the stream */
			g_mime_data_wrapper_write_to_stream ((GMimeDataWrapper*)content, stream);
			g_mime_stream_flush (stream);
			
			if (yenc != NULL)
			{
				/* log any crc errors */
				if ((yenc->pcrc!=0) && (yenc->pcrc != g_mime_filter_yenc_get_pcrc(yenc_filter)))
					Log::add_err_va (_("Checksum for `%s' part %d failed - file is corrupt"),
						         filename, part_number);
				if ((yenc->crc!=0) && (yenc->crc != g_mime_filter_yenc_get_crc(yenc_filter)))
					Log::add_err_va (_("Checksum for `%s' failed - file may be corrupt"),
						         filename);

				/* reset the filter to get ready for the next part */
				yenc_filter->pcrc = GMIME_YENCODE_CRC_INIT;
				yenc_filter->state = GMIME_YENCODE_STATE_INIT;
			}

			if (stream_file!=NULL && !ferror (GMIME_STREAM_FILE(stream_file)->fp))
				progress.set_status_va (_("Saved \"%s\""), filename);
			else {
				Log::add_err_va (_("Unable to save \"%s\" %s"),
					         filename,
                                                 file::pan_strerror(errno));
				unlink (filename);
				success = FALSE;
			}
		}

		g_object_unref (content);
	}

	if (success)
           Log::add_info_va (_("Saved \"%s\""), filename);

	/* cleanup */
	if (stream)
		g_object_unref (stream);
	stream = NULL;
	g_free (filename);
	filename = NULL;
	g_ptr_array_free (attachments, TRUE);

	return success;
}
